# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:50'

from urllib.parse import urljoin
from typing import Dict, Callable, Union
from sprite.http.headers import Headers
from sprite.exceptions import NotSupported
from sprite.http.request import Request
from sprite.http.cookies import CookiesJar


class Response:
    __slots__ = ('headers', 'status', '_body', '_url', '_cookiesJar', 'request', '_error')

    def __init__(self, url: str, status: int = 200, headers: Dict = None, body: str = "", cookies: Dict = None,
                 request: Request = None, error: Exception = None):
        self.headers = Headers(headers or {})
        self.status = int(status)
        self._body = body
        self._url = url
        self._cookiesJar = CookiesJar(cookies)
        self.request = request
        self._error = error

    @property
    def meta(self):
        try:
            return self.request.meta
        except AttributeError:
            raise AttributeError(
                "Response.meta not available, this response "
                "is not tied to any request"
            )

    @property
    def error(self) -> Union[Exception, None]:
        return self._error

    @property
    def url(self):
        return self._url

    @property
    def body(self):
        return self._body

    @property
    def cookies(self):
        return self._cookiesJar

    @property
    def callback(self):
        return self.request.callback

    def __str__(self):
        return "<%d %s>" % (self.status, self.url)

    __repr__ = __str__

    def copy(self):
        return self.replace()

    def replace(self, *args, **kwargs):
        for x in ['url', 'status', 'headers', 'body', 'request']:
            kwargs.setdefault(x, getattr(self, x))
        cls = kwargs.pop('cls', self.__class__)
        return cls(*args, **kwargs)

    def urljoin(self, url: str):
        return urljoin(self.url, url)

    def css(self, *a, **kw):
        raise NotSupported("Response content isn't text")

    def xpath(self, *a, **kw):
        raise NotSupported("Response content isn't text")

    def follow(self, url: str = None, callback: Callable = None, method: str = 'GET', headers: Dict = None,
               cookies: Dict = None, meta: Dict = None, encoding: str = 'utf-8', priority: int = 0,
               dont_filter: bool = False, ) -> Request:
        if url is None:
            url = self._url
        if headers is None:
            headers = self.headers
        if callback is None:
            callback = self.request.callback
        if cookies is None:
            cookies = self.request.cookies
        if meta is None:
            meta = self.request.meta
        return Request(url, callback,
                       method=method,
                       headers=headers,
                       cookies=cookies,
                       meta=meta,
                       encoding=encoding,
                       priority=priority,
                       dont_filter=dont_filter,
                       )



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/7/31 19:20'

from xmlrpc.server import SimpleXMLRPCServer
import socketserver
from xmlrpc.client import ServerProxy
import threading
import socket
from typing import Callable


# 远程调用默认重试次数
RETRY_TIMES = 10



class SpritRPCServer(socketserver.ThreadingMixIn, SimpleXMLRPCServer):
    def __init__(self, *args, **kwargs):
        SimpleXMLRPCServer.__init__(self, *args, **kwargs)
        self.allow_none = True
        self.allow_reuse_address = True

    # 注册远程方法
    def register_function(self, function:Callable, name:str=None, prefix:str=None):
        if prefix is not None:
            if name is None:
                name = function.__name__
            prefix = prefix + "_" if not prefix.endswith('_') else prefix
            SimpleXMLRPCServer.register_function(self, function, name=prefix + name)
        else:
            SimpleXMLRPCServer.register_function(self, function, name=name)


class ThreadSpriteRPCServer:
    def __init__(self, *args, **kwargs):
        self.rpc_server = SpritRPCServer(*args, **kwargs)
        self._t = threading.Thread(target=self.rpc_server.serve_forever)
        # self._t.setDaemon(True)
        self._t.start()

    def register_function(self, function:Callable, name:str=None, prefix:str=None):
        self.rpc_server.register_function(function, name=name, prefix=prefix)

    # 停止远程调用的服务
    def shutdown(self):
        self.rpc_server.shutdown()
        self._t.join()


# 发起远程调用
def client_call(server: str, func_name: str, *args, **kwargs):
    serv = ServerProxy(f'http://{server}')
    ignore = kwargs.get('ignore', False)
    # 调用失败是否重试
    if not ignore:
        err = None
        retry_times = 0
        while retry_times <= RETRY_TIMES:
            try:
                return getattr(serv, func_name)(*args)
            except socket.error as e:
                retry_times += 1
                err = e
            except Exception as e:
                err = e
                raise err
        raise err
    else:
        try:
            return getattr(serv, func_name)(*args)
        except socket.error as e:
            err = e
            raise err
        except Exception as e:
            err = e
            raise err



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-17 11:29'



from sprite.utils.pybloomfilter import BloomFilter



if __name__ == '__main__':
    b = BloomFilter(capacity=100)
    b.add("test")
    result = "test" in b
    result_tow = "test_two" in b
    print(result)
    print(result_tow)


# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-21 19:11'

import zlib
import inspect
import uuid
import os
from io import BytesIO


class GzipDecoder(object):

    def __init__(self):
        self._obj = zlib.decompressobj(16 + zlib.MAX_WBITS)

    def __getattr__(self, name):
        return getattr(self._obj, name)

    def decompress(self, data):
        if not data:
            return data
        return self._obj.decompress(data)

class BufferedIterable:
    def __init__(self, item):
        self.item = item
        self.cursor = self.item.__iter__()
        self.buffer = bytearray()

    def read(self, size):
        while len(self.buffer) < size:
            self.buffer.extend(self.cursor.__next__())
        temp = self.buffer[:size + 1]
        self.buffer = self.buffer[size + 1:]
        return temp

class FileUpload:
    def __init__(self, name: str=None, path: str=None, content: bytes=None, iterable=None,
                 f=None, headers: list=None):
        if not any([path, content, iterable, f]):
            raise Exception('You must supply either: path, content, iterable, f')
        self.name = name
        if f:
            self.f = f
        elif path:
            self.f = open(path, 'rb')
            if not self.name:
                self.name = os.path.basename(path)
        elif content:
            self.f = BytesIO(initial_bytes=content)
        elif iterable:
            self.f = BufferedIterable(iterable)
        if not self.name:
            self.name = str(uuid.uuid4())
        self.headers = headers
        self.is_async = inspect.iscoroutine(self.f.read)


class MultipartEncoder:

    def __init__(self, delimiter: bytes, params: dict, chunk_size: int=1*1024*1024,
                 loop=None, encoding: str='utf-8'):
        self.delimiter = b'--' + delimiter
        self.params = params
        self.chunk_size = chunk_size
        self.evaluated = False
        self.loop = loop
        self.encoding = encoding

    def create_headers(self, name: str, value) -> bytes:
        """

        :param name:
        :param value:
        :return:
        """
        if isinstance(value, FileUpload):
            return f'Content-Disposition: form-data; name="{name}"; filename="{value.name}"'.encode(self.encoding)
        else:
            return f'Content-Disposition: form-data; name="{name}"'.encode(self.encoding)

    def stream_value(self, value) -> bytes:
        """

        :param value:
        :return:
        """
        if isinstance(value, FileUpload):
            while True:
                if value.is_async:
                    chunk = self.loop.run_until_complete(value.f.read(self.chunk_size))
                else:
                    chunk = value.f.read(self.chunk_size)
                size = len(chunk)
                if size == 0:
                    break
                yield chunk
        else:
            if isinstance(value, int):
                yield str(value).encode()
            elif isinstance(value, str):
                yield value.encode(self.encoding)
            else:
                yield value

    def __iter__(self):
        """

        :return:
        """
        if self.evaluated:
            raise Exception('Streaming encoder cannot be evaluated twice.')
        for name, value in self.params.items():
            header = self.delimiter + b'\r\n' + self.create_headers(name, value) + b'\r\n\r\n'
            yield header
            for chunk in self.stream_value(value):
                yield chunk
            yield b'\r\n'
        yield self.delimiter + b'--'
        self.evaluated = True



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/7/31 19:27'


import socket


# 获取本机的ip地址
def get_ip():
    localIP = socket.gethostbyname(socket.gethostname())
    ex = socket.gethostbyname_ex(socket.gethostname())[2]
    if len(ex) == 1:
        return ex[0]
    for ip in ex:
        if ip != localIP:
            return ip


# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/7/3 19:28'

from typing import Coroutine, Callable, Union
import time
import asyncio
from asyncio import AbstractEventLoop
from threading import Thread, Event
from asyncio import Queue
from asyncio.locks import Lock
import traceback
from sprite.settings import Settings
from sprite.utils.log import get_logger

logger = get_logger()

"""
协程池
"""

# 默认最大协程数量
DefaultMaxCoroutinesAmount = 256 * 1024
# 默认每一个协程的空闲时间
DefaultMaxCoroutineIdleTime = 10

StopSingal = -1


class NotTooMangStartCorroutineException(Exception):
    pass


class NotStartCoroutineException(Exception):
    pass


class NotManyCoroutineException(Exception):
    pass


class CoroutineTask:
    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def task(coroutine_func: Callable):
        async def dosomething(*args, **kwargs):
            await coroutine_func(*args, **kwargs)

        return dosomething


class TaskQueue:
    def __init__(self):
        self._lastUseTime = time.time()
        self._queue = Queue()  # 协程的队列

    async def addTask(self, task: Coroutine):
        await self._queue.put(task)

    async def getTask(self) -> Coroutine:
        return await self._queue.get()

    def updateTime(self):
        self._lastUseTime = time.time()

    @property
    def lastUseTime(self) -> float:
        return self._lastUseTime


"""
每一个协程完成后，会自动加入准备队列，守护协程，会不断的检查这个队列，看准备队列里面的等待协程是否等待时候过长，等待时候过长的协程会被停掉
"""


class PyCoroutinePool:
    def __init__(self, maxCoroutineAmount: int = DefaultMaxCoroutinesAmount,
                 maxCoroutineIdleTime: int = DefaultMaxCoroutineIdleTime,
                 mostStop: bool = False, loop: AbstractEventLoop = None):
        self._maxCoroutineAmount = maxCoroutineAmount
        self._maxCoroutineIdleTime = maxCoroutineIdleTime
        self._mostStop = mostStop
        self._lock = Lock()
        self._ready = []
        self._loop = loop or asyncio.get_event_loop()
        self._stopEventLoopSingal = Event()
        self._stopEvent = None
        self._coroutinesCount = 0

    # 强制停止协程池,暂时不可用
    def force_stop(self):
        if self._stopEvent is None:
            raise NotStartCoroutineException("协程池还未开启")
        if self._stopEvent.is_set():
            logger.error("已经启动stop流程了，不用再发起了！！")
            return
        self._stopEvent.set()
        asyncio.run_coroutine_threadsafe(self._cancel_tasks(), self._loop)
        # self._stopEventLoopSingal.wait()
        # self._loop.close()

    async def _cancel_tasks(self):  # 取消任务
        tasks = []
        for task in asyncio.Task.all_tasks():  # 获取所有在事情循环中的task
            if not task.current_task() and not task.cancelled():
                tasks.append(task.cancel())  # 取消task
        # print("开始取消任务")
        await asyncio.gather(*tasks, return_exceptions=True)


    # 线程阻塞判断协程池时候已经停止运行
    def isStoped(self) -> bool:
        self._stopEventLoopSingal.wait()
        return True

    # 软关闭协程池
    def stop(self):
        if self._stopEvent is None:
            raise NotStartCoroutineException("协程池还未开启")
        if self._stopEvent.is_set():
            # logger.error("已经启动stop流程了，不用再发起了！！")
            return
        self._stopEvent.set()
        asyncio.run_coroutine_threadsafe(self._stop(), self._loop)
        asyncio.run_coroutine_threadsafe(self._stop_on_wait(), self._loop)

    async def _stop(self):
        async with self._lock:
            # 对于已经完成任务的协程，进行退出
            for taskQueue in self._ready:
                await taskQueue.addTask(None)
            # 没有完成的协程，提前发出通知，完成后，需要退出
            self._mostStop = True
            # print("发送信号完毕！！")

    async def _stop_on_wait(self):
        # while len(self._ready) > 0:
        #     asyncio.sleep(0.001)
        await self._cancel_tasks()
        self._loop.stop()
        self._stopEventLoopSingal.set()
        logger.info(f'协程池关闭')

    # 开始运行协程池，进行一些初始化工作
    def start(self, loop=None):
        if self._stopEvent is not None:
            raise NotTooMangStartCorroutineException("不能多次开启协程池")
        logger.info("开启协程池")
        self._stopEvent = Event()  # 创建事件
        # if self._loop is None:
        #     self._loop = loop or asyncio.get_event_loop()

        self._start()

    def _start(self):
        # asyncio.run_coroutine_threadsafe(self._daemon(), loop)
        self._init()  # 在子线程创建事件循环，并运行
        asyncio.run_coroutine_threadsafe(self._daemon(), self._loop)  # 开启协程

    def _init(self):
        run_loop_thread = Thread(
            target=self._start_subthread_loop, args=(self._loop,))  # 在子线程运行事件循环
        run_loop_thread.start()

    def _start_subthread_loop(self, loop):
        asyncio.set_event_loop(loop)
        # print("开启协程事件循环")
        loop.run_forever()  # 阻塞着一直运行事件循环，直到被调用stop    loop.stop()
        # print("退出事件循环监听！！")

    async def _clean(self):
        scratch = []
        maxCoroutineIdleTime = self._maxCoroutineIdleTime
        currentTime = time.time()
        async with self._lock:
            n = len(self._ready)
            i = 0
            while i < n:
                # 比较时间
                if currentTime - self._ready[i].lastUseTime > maxCoroutineIdleTime:
                    i += 1
                else:
                    break
            if i > 0:
                # 选出没有空闲的协程,抛弃空闲过久的协程
                scratch = self._ready[:i]
                self._ready = self._ready[i:]  # 第i位不满足
                # for index in range(0, i):
                #     self._ready.pop(index)
            # 停掉空闲过久的协程
            for taskQueue in scratch:
                await taskQueue.addTask(None)
        # print("清理一次完毕")

    async def _daemon(self):
        # 协程进行守护工作
        # print("开启更新协程")
        while not self._stopEvent.is_set():
            await self._clean()
            await asyncio.sleep(self._maxCoroutineIdleTime)
        # print("退出清理任务")

    # 添加任务执行
    def go(self, task: Coroutine) -> bool:
        try:
            asyncio.run_coroutine_threadsafe(
                self._goCoroutine(task), self._loop)
            return True
        except NotManyCoroutineException as e:
            print(e)
            return False

    async def _goCoroutine(self, task: Coroutine):
        taskQueue = await self._getIdleCoroutine()
        # print("获取到一个任务队列实例")
        if taskQueue is None:
            # print("发生错误")
            raise NotManyCoroutineException("没有更多协程了")

        await taskQueue.addTask(task)
        # print("向一个任务队列实例添加任务成功")

    # 获取空闲协程，不存在就新创建一个
    async def _getIdleCoroutine(self) -> Union[TaskQueue, None]:
        taskQueue = None
        createCoroutine = False
        async with self._lock:
            n = len(self._ready) - 1
            if n < 0:
                if self._coroutinesCount < self._maxCoroutineAmount:
                    createCoroutine = True
                    self._coroutinesCount += 1
            else:
                # taskQueue = self._ready[n]
                # self._ready.pop(n)  # 去掉这个task

                taskQueue = self._ready.pop(n)  # 去掉这个task
        if taskQueue is None:
            if not createCoroutine:
                return None
            taskQueue = TaskQueue()
            # await self._executeTask(newTaskQueue) # 在协程中执行任务
            # 新开启一个协程并让这个协程进入准备阶段
            asyncio.run_coroutine_threadsafe(
                self._executeTask(taskQueue), self._loop)
        return taskQueue

    async def _executeTask(self, taskQueue):
        while True:
            # print("开始接受协程任务！！！")
            task = await taskQueue.getTask()  # 阻塞等待任务
            if task is None:
                # 接收到停止信号
                # print("接收到停止信号！")
                break
            # 执行任务
            # print("接收到一个协程任务")
            try:
                await task
            except:
                logger.error("find one error: ")
                traceback.print_exc()
            # 每一个协程完成任务后，是否需要停止这个协程
            if await self._isRelease(taskQueue):
                # 停掉目前的协程
                # print("接收到stop的命令")
                break
        async with self._lock:
            # print("正在运行的协程数量—1")
            self._coroutinesCount -= 1

    async def _isRelease(self, taskQueue: TaskQueue) -> bool:
        taskQueue.updateTime()  # 更新任务完成时间
        async with self._lock:
            if self._mostStop:
                return True
            # 任务完成的协程就加入到准备队列里面
            self._ready.append(taskQueue)
            return False

    @classmethod
    def from_setting(cls, settings: Settings):
        maxCoroutineAmount = settings.getint("MAXCOROUTINEAMOUNT")
        maxCoroutineIdleTime = settings.getint("MAXCOROUTINEIDLETIME")
        mostStop = settings.getbool("MOSTSTOP")
        obj = cls(maxCoroutineAmount=maxCoroutineAmount,
                  maxCoroutineIdleTime=maxCoroutineIdleTime,
                  mostStop=mostStop)
        return obj

    @property
    def loop(self):
        return self._loop


if __name__ == '__main__':
    @CoroutineTask.task
    async def sleepTask(task_id: int, sleep_time: int):
        print(f'在协程中启动任务 {task_id}')
        await asyncio.sleep(sleep_time)
        a = [1]
        print(a[9])
        print(f'{task_id} 任务结束')

    async def stop_corountine(coroutine_pool: PyCoroutinePool):
        coroutine_pool.stop()

    coroutine_pool = PyCoroutinePool()
    coroutine_pool.start()
    for index in range(10):
        coroutine_pool.go(sleepTask(task_id=index, sleep_time=index))

    time.sleep(3)
    print("主线程发出stop信号")
    # coroutine_pool.go(stop_corountine(coroutine_pool))
    coroutine_pool.stop()



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-21 13:49'

import httptools
import time
import asyncio
from asyncio import wait_for
import json
from enum import Enum
from .request import Request
from .connection import Connection
from .headers import Headers
from sprite.http.cookies import CookiesJar, Cookie
from sprite.utils.decoders import GzipDecoder
from sprite.exceptions import StreamAlreadyConsumed
from .exceptions import InvalidResponseException


class ResponseStatus(Enum):
    PENDING_HEADERS = 1
    PENDING_BODY = 2
    CHUNKED_TRANSFER = 3
    DECODING = 4
    COMPLETE = 5
    ERROR = 6


class Response:
    __slots__ = ('_connection', '_headers', '_content', '_parser', '_parser_status', '_cookies',
                 '_status_code', '_decode', '_chunk_size', '_decoder', '_encoding',
                 'request', 'url')

    def __init__(self, url: str, connection: Connection,
                 request: Request, chunk_size: int = 1 * 1024 * 1024, decode: bool = True):
        self._connection = connection
        self._headers = Headers()
        self._content = bytearray()
        # noinspection PyArgumentList
        # 调用C的http协议的解析器，解析http
        # self._parser = HttpResponseParser(self)
        self._parser = httptools.HttpResponseParser(self)
        # 用来协调http协议的数据的解析进度，默认是从解析headers状态开始
        self._parser_status = ResponseStatus.PENDING_HEADERS
        self._cookies = None
        self._status_code = None
        self._chunk_size = chunk_size
        self._decode = decode
        self._decoder = None
        self._encoding = None
        self.request = request
        self.url = url

    @property
    def encoding(self):
        return 'utf-8'

    def json(self, *args, loads=None, **kwargs):
        if not loads:
            loads = json.loads
        return loads(self.content.decode(self.encoding), *args, **kwargs)

    def text(self, encoding: str = None) -> str:
        return self.content.decode(encoding=encoding or self.encoding)

    @property
    def status_code(self):
        """

        :return:
        """
        if self._parser_status == ResponseStatus.PENDING_HEADERS:
            raise Exception('Status code not loaded yet. '
                            'In streaming mode you should manually call load_headers().')
        return self._status_code

    @property
    def headers(self):
        """

        :return:
        """
        if self._parser_status == ResponseStatus.PENDING_HEADERS:
            raise Exception('Headers not loaded yet. '
                            'In streaming mode you should manually call load_headers().')
        # self._headers.eval()
        # tmp_headers = {}
        # print(self._headers.values)
        # for key, value in self._headers.items():
        #     tmp_headers[key.decode("utf-8")] = value.decode("utf-8")
        return self._headers

    @property
    def content(self):
        """

        :return:
        """
        if self._parser_status == ResponseStatus.PENDING_BODY:
            raise Exception('You need to call read_content() '
                            'before using this in streaming mode.')
        return self._content

    @content.setter
    def content(self, data: bytearray):
        self._content = data

    @property
    async def cookies(self):
        """

        :return:
        """
        # 目前的解析进度是读取并解析headers
        if self._parser_status == ResponseStatus.PENDING_HEADERS:
            await self.receive_headers()
        if self._cookies is None:
            self._cookies = CookiesJar()
            # headers里面有set-cookie的键值对
            if self._headers.get('set-cookie'):
                # 从headers里面读取set-cookie键的值，并解析后，存储在cookies实例中
                self._cookies.add_cookie(Cookie.from_header(self._headers['set-cookie']))
        return self._cookies

    async def read_content(self):
        # 读取完headers之后，开始根据headers提供的元信息读取body
        """

        :return:
        """
        await self.receive_headers()
        if self._parser_status != ResponseStatus.PENDING_BODY:
            return
        try:
            length = int(self._headers['Content-Length'])
            if not self._decoder:
                # Skipping the HTTP parser for performance.
                self._content.extend(await self._connection.read_exactly(length))
                self.on_message_complete()
            else:
                self._parser.feed_data(await self._connection.read_exactly(length))
        except KeyError:
            if self._headers.get('Transfer-Encoding') == 'chunked':
                await self._handle_chunked_encoding()
            else:
                raise InvalidResponseException('Invalid response.')
            return self._content
        except ValueError:
            raise Exception('Invalid content-length header.')

    async def _handle_chunked_encoding(self):
        """

        :return:
        """
        self._parser_status = ResponseStatus.CHUNKED_TRANSFER
        while self._parser_status == ResponseStatus.CHUNKED_TRANSFER:
            self._parser.feed_data(await self._connection.read_until(b'\r\n'))

    async def receive_headers(self):
        # 开始接受头部数据并解析
        """

        :return:
        """
        # 目前的解析进度是开始解析headers
        if self._parser_status != ResponseStatus.PENDING_HEADERS:
            return
        try:
            # 从connection中读取到指定字符的数据，并把读取的数据放入解析器数据流方法中
            resp_header = await self._connection.read_until(b'\r\n\r\n')
            # print(f'befor feed _headers: {self._headers}')
            # print(f'resp_header {resp_header}')
            self._parser.feed_data(resp_header)
            # print(f'after feed _headers: {self._headers}')
            # print(f'after feed headers type {type(self._headers)}')
            if self._headers.get('content-encoding') == 'gzip':
                self._decoder = GzipDecoder()
            # 解析进度更新为解析读取body并解析body
            self._parser_status = ResponseStatus.PENDING_BODY
        except OSError as error:
            # In case any network error occurs we should discard this connection.
            self._parser_status = ResponseStatus.ERROR
            self._connection.close()
            raise error

    async def _release_connection(self):
        """

        :return:
        """
        await self._connection.pool.release_connection(self._connection, self._parser.should_keep_alive())

    async def stream(self, chunk_size: int = 1 * 1024 * 1024, chunk_timeout: int = 10,
                     complete_timeout: int = 300):
        # 读取流式数据
        """

        :param complete_timeout:
        :param chunk_timeout:
        :param chunk_size:
        :return:
        """
        if self._parser_status not in (ResponseStatus.PENDING_HEADERS, ResponseStatus.PENDING_BODY):
            raise StreamAlreadyConsumed
        if self._parser_status == ResponseStatus.PENDING_HEADERS:
            await wait_for(self.receive_headers(), chunk_timeout)

        try:
            # 尝试直接接受指定长度的内容
            length = int(self.headers['Content-Length'])
            remaining = length
            while remaining:
                # 确定尝试读取的字节长度
                bytes_to_read = min(remaining, chunk_size)
                task = self._connection.read_exactly(bytes_to_read)
                start_time = time.time()
                self._parser.feed_data(await wait_for(task, min(chunk_timeout, complete_timeout)))
                complete_timeout -= time.time() - start_time
                remaining -= bytes_to_read
                yield bytes(self._content)
                self._content = bytearray()
        except KeyError:
            # 内容采用分块传输的方式
            if self._headers.get('Transfer-Encoding') == 'chunked':
                self._parser_status = ResponseStatus.CHUNKED_TRANSFER
                while self._parser_status == ResponseStatus.CHUNKED_TRANSFER:
                    task = self._connection.read_until(b'\r\n')
                    start_time = time.time()
                    try:
                        self._parser.feed_data(await wait_for(task, min(chunk_timeout, complete_timeout)))
                    except asyncio.LimitOverrunError as error:
                        self._parser.feed_data(await self._connection.read_exactly(error.consumed))
                    complete_timeout -= time.time() - start_time
                    while len(self._content) >= chunk_size:
                        yield self._content[:chunk_size]
                        self._content = self._content[chunk_size:]
                while len(self._content) > 0:
                    yield self._content[:chunk_size]
                    self._content = self._content[chunk_size:]
            else:
                raise Exception('Invalid response.')
        except ValueError:
            raise Exception('Invalid content-length header.')

    def is_redirect(self):
        return self.headers.get('location') is not None

    #############
    # # #  Http Parser Callbacks
    ######################################

    def on_message_begin(self):
        pass

    def on_url(self, url: bytes):
        pass

    def on_header(self, name: bytes, value: bytes):
        """

        :param name:
        :param value:
        :return:
        """
        self._headers.raw.append((name, value))

    def on_headers_complete(self):
        pass

    def on_body(self, data: bytes):
        """

        :param data:
        :return:
        """
        if self._decoder is not None:
            self._content.extend(self._decoder.decompress(data))
        else:
            self._content.extend(data)

    def on_message_complete(self):
        """

        :return:
        """
        self._parser_status = ResponseStatus.COMPLETE
        try:
            self._connection.loop.create_task(self._release_connection())
        except Exception as error:
            print(error)

    def on_chunk_header(self):
        pass

    def on_chunk_complete(self):
        pass

    def on_status(self, status: bytes):
        self._status_code = status.decode('utf-8') if status.decode('utf-8') == "ok" else 200

    def chunk_complete(self):
        pass

    def __repr__(self):
        if 400 > self.status_code >= 300:
            if len(self.url) > 30:
                url = self.url[:30] + '...'
            else:
                url = self.url
            return f'<Response [{self.status_code} => "{url}"]>'
        else:
            return f'<Response [{self.status_code}]>'


class BodyStream:
    def __init__(self, response: Response):
        self.response = response

    async def __aiter__(self):
        return self

    async def __anext__(self):
        pass



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-25 22:09'

from urllib import parse


class URL:

    def __init__(self, schema: bytes, host: bytes, port, path: bytes,
                 query: bytes, fragment: bytes, userinfo: bytes):
        self.schema = schema.decode('utf-8')
        self.host = host.decode('utf-8')
        self.port = port
        self.path = path.decode('utf-8')
        self.query = query.decode('utf-8')
        self.fragment = fragment.decode('utf-8')
        self.userinfo = userinfo.decode('utf-8')
        self.netloc = self._setNetloc()

    def __repr__(self):
        return ('<URL schema: {!r}, host: {!r}, port: {!r}, path: {!r}, '
                'query: {!r}, fragment: {!r}, userinfo: {!r}>'
                .format(self.schema, self.host, self.port, self.path, self.query, self.fragment, self.userinfo))

    @property
    def raw(self):
        _raw = self.netloc
        if self.path:
            _raw += '/' + self.path
            if self.query:
                _raw += '?' + self.query
        return _raw

    def _setNetloc(self):
        if self.query:
            return self.path + "?" + self.query
        return self.path


def parse_url(url: bytes) -> URL:
    parse_result = parse.urlparse(url)
    return URL(schema=parse_result.scheme,
               host=parse_result.netloc.split(b':')[0],
               port=parse_result.port or 0,
               path=parse_result.path,
               query=parse_result.query,
               fragment=parse_result.fragment,
               userinfo=b'')



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:45'

import time
import traceback
import asyncio
import threading
from typing import Callable, Coroutine
from types import AsyncGeneratorType
from sprite.core.scheduler import Slot, Scheduler
from sprite.core.download import Downloader
from sprite.utils.coroutinePool import PyCoroutinePool
from sprite.middlewaremanager import MiddlewareManager
from sprite.settings import Settings
from sprite.spider import Spider
from sprite.http.request import Request
from sprite.http.response import Response
from sprite.utils.log import get_logger
from sprite.item import Item
from sprite.utils.request import Counter
from sprite.utils.asyncHandler import detailCallable
from sprite.exceptions import SchedulerEmptyException

logger = get_logger()


class Engine:
    def __init__(self, scheduler: Scheduler, downloader: Downloader, coroutine_pool: PyCoroutinePool,
                 middlewareManager: MiddlewareManager,
                 spider: Spider, settings: Settings):
        self._slot = Slot()
        self._scheduler = scheduler
        self._downloader = downloader
        self._coroutine_pool = coroutine_pool
        self._middlewareManager = middlewareManager
        self._spider = spider
        self._settings = settings

        self._running = threading.Event()
        self._init = threading.Event()
        self._unfinished_workers = 0

        self._start_time = time.time()
        self._downloaded_request_count = 0
        self._success_request_count = 0
        self._failed_request_count = 0

        self._item_counter = Counter(unit=settings.getint("ITEM_COUNTER_UNIT"))
        self._response_counter = Counter(
            unit=settings.getint("RESPONSE_COUNTER_UNIT"))

    def isClose(self) -> bool:
        if self._running.is_set():
            return True
        return False

    def start(self):
        logger.info(f'启动engine')
        # 启动调度器
        self._scheduler.start()
        # 启动协程池
        self._coroutine_pool.start()
        # 注入start_requests
        self._coroutine_pool.go(self._get_start_requests())

        # 查询调度器中是否填充了request，如果没有直接退出程序
        self._init.wait()
        if not self._scheduler.has_pending_requests():
            logger.error(
                "scheduler 为空，没有构造start request或者填充start request失败, 关闭程序")
            self.close()
            return

        # 执行爬虫中间件
        self._coroutine_pool.go(
            self._middlewareManager.process_spider_start(self._spider))
        # 设定工作协程的数量
        self._unfinished_workers = self._settings.getint("WORKER_NUM")

        # 检测工作协程是否都退出
        self._coroutine_pool.go(self._status_check())
        # 启动所有的工作协程
        for _ in range(self._unfinished_workers):
            self._coroutine_pool.go(self._doSomething())

    async def _get_start_requests(self):
        if not self._scheduler.has_pending_requests():
            # 检测非断点续爬
            try:
                if self._spider.start_requests is not None:
                    for url in self._spider.start_requests:
                        request = Request(url=url, headers=self._settings.getdict(
                            "HEADERS"), callback=self._spider.parse)
                        self._scheduler.enqueue_request(request)
                else:
                    # async for request in self._spider.start_request():
                    #     self._scheduler.enqueue_request(request)
                    await detailCallable(self._spider.start_request, self._scheduler.enqueue_request)
            except Exception as e:
                logger.info(
                    f'填充start request 过程中发生错误: \n{traceback.format_exc()}')
        self._init.set()

    # 不断检测所有的工作协程是否都结束，如果都结束的化，则启动关闭引擎的流程
    async def _status_check(self):
        while True:
            if self._unfinished_workers <= 0:
                self.close()
                break
            await asyncio.sleep(0.001)

    async def _doSomething(self):
        while True:
            # 1.首先判断spider非close
            if not self._running.is_set():
                # 2.再尝试从调度器里面提取request
                try:
                    # 获取到request之后，开始处理请求,先将request放入正在处理队列记录一下
                    self._slot.addRequest(self._scheduler.next_request())
                    # 再取出
                    request = self._slot.getRequest()
                    try:
                        await self._doCrawl(request)
                    except Exception as e:
                        logger.error(
                            f'find one error: \n{traceback.format_exc()}')
                    # 处理完一个request，打一个标记
                    self._slot.toDone()
                except SchedulerEmptyException:
                    if self._slot.has_pending_request():
                        await asyncio.sleep(0.001)
                        continue
                    else:
                        self._unfinished_workers -= 1
                        break
            else:
                self._unfinished_workers -= 1
                break

    # 获取到request之后，开始处理请求
    async def _doCrawl(self, request: Request):
        response = None
        # 1.首先调用下载中间件
        result = await self._middlewareManager.process_request(request, spider=self._spider)
        if result:
            if isinstance(result, Request):
                request = result
            elif isinstance(result, Response):
                response = result
        if response is None:
            # 2.调用下载器下载request
            logger.debug(f'downloading request: {request.url} {request.query}')
            response = await self._downloader.request(request=request)
            self._downloaded_request_count += 1
            if response.error:
                self._failed_request_count += 1
                logger.debug(
                    f'downloaded request failure: {request.url} {request.query}')
            else:
                logger.debug(
                    f'downloaded request: {request.url}[{response.status}]'
                )
                unit_speed, unit_count = self._response_counter.dot()
                if unit_speed or unit_count:
                    logger.info(
                        f'目前获取response的速度：{unit_speed}/s   每{self._item_counter.unit}s获取{unit_count}个response')
                self._success_request_count += 1
        # 3.调用下载中间件处理response
        result = await self._middlewareManager.process_response(response, self._spider)
        if result:
            if isinstance(result, Request):
                # 丢入调度器中
                request = result
                self._scheduler.enqueue_request(request)
                return
            elif isinstance(result, Response):
                response = result
        # 4.调用request的回调函数，对response进行处理
        await self._handle_request_callback(request.callback, response)

    async def _handle_request_callback(self, callback: Callable, response: Response):
        # 传入resposne，调用回调函数进行处理
        # 调用回调函数
        callback_results = callback(response)
        if isinstance(callback_results, AsyncGeneratorType):
            # print(f'回调函数是 AsyncGeneratorType')
            return await self._process_async_callback(callback_results=callback_results)
        elif isinstance(callback_results, Coroutine):
            # print(f'回调函数是 一个Coroutine')
            result = await self._handle_coroutine_callback(callback_results)
            if result:
                if isinstance(result, Request):
                    self._scheduler.enqueue_request(result)
                elif isinstance(result, Item):
                    # 5. 调用管道处理item
                    await self._middlewareManager.pipeline_process_item(item=result, spider=self._spider)

    # 处理回调（以协程的方式）
    async def _process_async_callback(self, callback_results: AsyncGeneratorType):
        async for callback_result in callback_results:  # AsyncGeneratorType对象 协程生成器对象
            # yield 的返回结果类型
            if isinstance(callback_result, AsyncGeneratorType):
                # 继续递归
                await self._process_async_callback(callback_result)
            elif isinstance(callback_result, (Request, Item)):
                # yield 的返回值是Request or Item类型
                if callback_result:
                    if isinstance(callback_result, Request):
                        self._scheduler.enqueue_request(callback_result)
                    elif isinstance(callback_result, Item):
                        logger.debug(
                            f'crawled one Item: {callback_result}'
                        )
                        # 5. 调用管道处理item
                        unit_speed, unit_count = self._item_counter.dot()
                        if unit_count or unit_speed:
                            logger.info(
                                f'目前获取item的速度：{unit_speed}/s   每{self._item_counter.unit}s获取{unit_count}个item')
                        await self._middlewareManager.pipeline_process_item(item=callback_result, spider=self._spider)
            elif isinstance(callback_result, Coroutine):
                await self._handle_coroutine_callback(callback_result)

    async def _handle_coroutine_callback(self, aws_callback: Coroutine):
        result = await aws_callback
        if result:
            if isinstance(result, Request):
                self._scheduler.enqueue_request(result)
            elif isinstance(result, Item):
                # 5. 调用管道处理item
                await self._middlewareManager.pipeline_process_item(item=result, spider=self._spider)

    def get_download_staticate(self) -> [int, int, int]:
        return self._failed_request_count, self._success_request_count, self._downloaded_request_count

    # 对外的接口，用于关闭引擎
    def close(self):
        if self._running.is_set():
            return
        # 1. 首先关闭发出关闭信号，
        self._running.set()
        # 2.在协程中执行关闭流程
        self._coroutine_pool.go(self._close())

    async def _close(self):
        # 关闭引擎的步骤
        # 1.首先关闭调度器
        self._scheduler.close()
        # 2.等待正在执行的reques执行结束
        await self._slot.join()
        logger.info("正在处理的request处理完毕！")
        logger.info(f'一共耗时：{time.time() - self._start_time}s')
        logger.info(
            f'下载情况统计： 一共发送请求：{self._downloaded_request_count}    成功请求数量：{self._success_request_count}    失败请求数量：{self._failed_request_count}')
        # 3.关闭所有的tcp连接
        self._downloader.close()
        # 4.然后关闭协程池
        await self._middlewareManager.process_spider_close(self._spider)

        # 5.执行爬虫中间件
        self._coroutine_pool.stop()

    @classmethod
    def from_settings(cls, settings: Settings, spider: Spider, middlewareManager: MiddlewareManager):
        if middlewareManager is None:
            middlewareManager = MiddlewareManager()
        coroutine_pool = PyCoroutinePool.from_setting(settings)
        obj = cls(
            scheduler=Scheduler.from_settings(settings, spider),
            downloader=Downloader.from_settings(settings, coroutine_pool),
            coroutine_pool=coroutine_pool,
            spider=spider,
            settings=settings,
            middlewareManager=middlewareManager,
        )
        return obj



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:58'

import collections
from collections import defaultdict, deque
from itertools import chain
from asyncio import coroutine, events, locks, QueueEmpty, QueueFull



# 实现优先队列
# deque 双向队列
class PriorityQueue:
    def __init__(self):
        self.negitems = defaultdict(deque)
        self.pzero = deque()
        self.positems = defaultdict(deque)

    def push(self, item, priority:int=0):  # 压入元素
        # 同样的权重，需按先进先出的原则来，不同的权重按权重大小来
        if priority == 0:
            self.pzero.appendleft(item)
        elif priority < 0:  # 权重大于或者小于0
            self.negitems[priority].appendleft(item)  # 权重小于0
        else:
            self.positems[priority].appendleft(item)  # 权重大于0

    def pop(self):  # 弹出元素
        # 弹出顺序:  先从权重小于0的开始选择
        #           再从权重等于0的开始选择
        #           最后从权重大于0的开始选择
        if self.negitems:
            priorities = list(self.negitems.keys())
            priorities.sort()  # 默认是升序排序，
            for priority in priorities:  # 从最小的权重开始遍历，看对应的双端队列是否为空
                deq = self.negitems[priority]
                if deq:  # 不为空，就弹出一个元素
                    t = (deq.pop(), priority)
                    if not deq:
                        del self.negitems[priority]  # 弹出元素为空后，删除掉双端队列
                    return t  # 返回值是由 值和权重组成的元组
        elif self.pzero:
            return (self.pzero.pop(), 0)
        else:
            priorities = list(self.positems.keys())
            priorities.sort()
            for priority in priorities:
                deq = self.positems[priority]
                if deq:
                    t = (deq.pop(), priority)
                    if not deq:
                        del self.positems[priority]
                    return t
        raise IndexError("pop from an empty queue")  # 优先队列为空

    def __len__(self):
        total = sum(len(v) for v in self.negitems.values()) + \
                len(self.pzero) + \
                sum(len(v) for v in self.positems.values())
        return total

    def __iter__(self):  # 优先队列遍历方法
        gen_negs = ((i, priority)
                    for priority in sorted(self.negitems.keys())
                    for i in reversed(self.negitems[priority]))
        gen_zeros = ((item, 0) for item in self.pzero)
        gen_pos = ((i, priority)
                   for priority in sorted(self.positems.keys())
                   for i in reversed(self.positems[priority]))
        return chain(gen_negs, gen_zeros, gen_pos)

    def __nonzero__(self):
        return bool(self.negitems or self.pzero or self.positems)




class Queue:
    """
    对标准库中的协程队列的细微改造
    """

    def __init__(self, maxsize=0, *, loop=None):
        if loop is None:
            self._loop = events.get_event_loop()
        else:
            self._loop = loop
        self._maxsize = maxsize

        # Futures.
        self._getters = collections.deque()
        # Futures.
        self._putters = collections.deque()
        self._unfinished_tasks = 0
        self._finished = locks.Event(loop=self._loop)
        self._finished.set()
        self._init(maxsize)


    def _init(self, maxsize):
        self._queue = collections.deque()

    def _get(self):
        return self._queue.popleft()

    def _put(self, item):
        self._queue.append(item)


    def _wakeup_next(self, waiters):
        while waiters:
            waiter = waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)
                break

    def __repr__(self):
        return '<{} at {:#x} {}>'.format(
            type(self).__name__, id(self), self._format())

    def __str__(self):
        return '<{} {}>'.format(type(self).__name__, self._format())

    def _format(self):
        result = 'maxsize={!r}'.format(self._maxsize)
        if getattr(self, '_queue', None):
            result += ' _queue={!r}'.format(list(self._queue))
        if self._getters:
            result += ' _getters[{}]'.format(len(self._getters))
        if self._putters:
            result += ' _putters[{}]'.format(len(self._putters))
        if self._unfinished_tasks:
            result += ' tasks={}'.format(self._unfinished_tasks)
        return result

    def qsize(self):
        return len(self._queue)

    @property
    def maxsize(self):
        return self._maxsize

    def empty(self):
        if self._unfinished_tasks>0:
            return False
        return True

    def full(self):
        if self._maxsize <= 0:
            return False
        else:
            return self.qsize() >= self._maxsize

    @coroutine
    def put(self, item):
        while self.full():
            putter = self._loop.create_future()
            self._putters.append(putter)
            try:
                yield from putter
            except:
                putter.cancel()  # Just in case putter is not done yet.
                if not self.full() and not putter.cancelled():
                    # We were woken up by get_nowait(), but can't take
                    # the call.  Wake up the next in line.
                    self._wakeup_next(self._putters)
                raise
        return self.put_nowait(item)

    def put_nowait(self, item):
        if self.full():
            raise QueueFull
        self._put(item)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    @coroutine
    def get(self):
        while self.empty():
            getter = self._loop.create_future()
            self._getters.append(getter)
            try:
                yield from getter
            except:
                getter.cancel()  # Just in case getter is not done yet.

                try:
                    self._getters.remove(getter)
                except ValueError:
                    pass

                if not self.empty() and not getter.cancelled():
                    # We were woken up by put_nowait(), but can't take
                    # the call.  Wake up the next in line.
                    self._wakeup_next(self._getters)
                raise
        return self.get_nowait()

    def get_nowait(self):
        if self.empty():
            raise QueueEmpty
        item = self._get()
        self._wakeup_next(self._putters)
        return item

    def task_done(self):
        if self._unfinished_tasks <= 0:
            raise ValueError('task_done() called too many times')
        self._unfinished_tasks -= 1
        if self._unfinished_tasks == 0:
            self._finished.set()

    @coroutine
    def join(self):
        if self._unfinished_tasks > 0:
            yield from self._finished.wait()


if __name__ == '__main__':
    



    test_queue = Queue()
    test_queue.put_nowait("test")
    print(test_queue.empty())
    print("========================")
    test_queue.get_nowait()
    print(test_queue.empty())
    print("========================")
    test_queue.task_done()
    print(test_queue.empty())







# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:46'


from typing import Callable, List, Coroutine, Awaitable
from sprite.http.request import Request
from sprite.http.response import Response
from sprite.item import Item
from sprite.spider import Spider

from sprite.exceptions import TypeNotSupport


DOWNLOAD_MIDDLEWARE, PIPELINE_MIDDLEWARE, SPIDER_MIDDLEWARE = range(3)

BEFORE = "before"
AFTER = "after"


class MiddlewareManager:
    def __init__(self):
        self._middlewares = {

        DOWNLOAD_MIDDLEWARE: {
            BEFORE: [],
            AFTER: [],
        },

        PIPELINE_MIDDLEWARE: [],

        SPIDER_MIDDLEWARE: {
            BEFORE: [],
            AFTER: [],
        },
    }

    def _type_check(self, func:Callable):
        if not isinstance(func, Awaitable):
            raise TypeNotSupport("中间件类型必须是协程！")

    def add_download_middleware(self, isBefore: bool = True):
        def wrappFunc(func: Callable):
            def register_middleware(func: Callable):
                # self._type_check(func)
                if isBefore:
                    self._middlewares[DOWNLOAD_MIDDLEWARE][BEFORE].append(func)
                else:
                    self._middlewares[DOWNLOAD_MIDDLEWARE][AFTER].append(func)

            register_middleware(func)

        return wrappFunc

    def add_pipeline_middleware(self):
        def wrappFunc(func: Callable):
            def register_middleware(func):
                # self._type_check(func)
                self._middlewares[PIPELINE_MIDDLEWARE].append(func)

            register_middleware(func)

        return wrappFunc

    def add_spider_middleware(self, isBefore: bool = True):
        def wrappFunc(func: Callable):
            def register_middleware(func: Callable):
                # self._type_check(func)
                if isBefore:
                    self._middlewares[SPIDER_MIDDLEWARE][BEFORE].append(func)
                else:
                    self._middlewares[SPIDER_MIDDLEWARE][AFTER].append(func)

            register_middleware(func)

        return wrappFunc

    def _getMiddleWareByPostion(self, middleware_name: str, postion: str=None) -> List:
        if postion:
            return self._middlewares[middleware_name][postion]
        return self._middlewares[middleware_name]


    async def process_request(self,request:Request, spider:Spider):
        fname = lambda f: f'{f.__name__}'
        result = None
        for method in self._getMiddleWareByPostion(middleware_name=DOWNLOAD_MIDDLEWARE,postion=BEFORE):
            result =await method(request=request, spider=spider)
            assert result is None or isinstance(result, (Request, Response)), f'this middleware {fname(method)} must return None or Resquest or Response, but got {type(result)}'
            # 有返回值表示 下面的中间件不再执行
            if result:
                return result
        return result

    async def process_response(self, response:Response , spider:Spider):
        fname = lambda f: f'{f.__name__}'
        result = None
        for method in self._getMiddleWareByPostion(middleware_name=DOWNLOAD_MIDDLEWARE, postion=AFTER):
            result =await method(response=response, spider=spider)
            assert result is None or isinstance(result, (Request, Response)), f'this middleware {fname(method)} must return None or Resquest or Response, but got {type(result)}'
            if result:
                return result
        return result


    async def pipeline_process_item(self, item:Item, spider:Spider):
        fname = lambda f:f'{f.__name__}'
        for method in self._getMiddleWareByPostion(middleware_name=PIPELINE_MIDDLEWARE):
            result =await method(spider=spider, item=item)
            assert result is None or isinstance(result, Item), f'this middleware {fname(method)} must return None or Item, but got {type(result)}'
            if result is None:
                return

    async def process_spider_start(self, spider:Spider):
        for method in self._getMiddleWareByPostion(middleware_name=SPIDER_MIDDLEWARE, postion=BEFORE):
            await method(spider=spider)


    async def process_spider_close(self, spider:Spider):
        for method in self._getMiddleWareByPostion(middleware_name=SPIDER_MIDDLEWARE, postion=AFTER):
            await method(spider=spider)





# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-21 13:49'

from .url import URL
from sprite.http.cookies import CookiesJar
from .connection import Connection


class Request:
    __slots__ = ('method', 'url', 'headers', 'data', 'cookies', 'streaming', 'chunked',
                 'encoding', 'origin')

    def __init__(self, method: str, url: URL, headers: dict, data, cookies: CookiesJar, origin=None):
        self.method = method.upper() if method else 'GET'
        self.url = url
        self.headers = headers or {}
        self.data = data or b''
        self.cookies = cookies
        self.streaming = True if data and not isinstance(
            data, (bytes, bytearray)) else False
        self.chunked = self.streaming
        self.origin = origin

    async def encode(self, connection: Connection):
        # 把请求编码成字节类型，并调用connection实例发送请求
        http_request = f'{self.method} {self.url.netloc} HTTP/1.1\r\n'
        # Headers
        for header, value in self.headers.items():
            http_request += f'{header}: {str(value)}\r\n'
        # Cookies
        if self.cookies:
            cookies = ';'.join([c.name + '=' + c.value for c in self.cookies])
            http_request += f'Cookie: {cookies}'

        if not self.streaming:
            http_request += f'Content-Length: {str(len(self.data))}\r\n'
            await connection.sendall((http_request + '\r\n').encode() + self.data)

        elif self.chunked:
            http_request += f'Transfer-Encoding: chunked\r\n'
            await connection.sendall((http_request + '\r\n').encode())
            for chunk in self.data:
                size = hex(len(chunk))[2:].encode('utf-8')
                await connection.sendall(size + b'\r\n' + chunk + b'\r\n')
            await connection.sendall(b'0\r\n\r\n')



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/19 16:41'


from sprite.middlewaremanager import MiddlewareManager
from sprite.middlewaremanager import DOWNLOAD_MIDDLEWARE, BEFORE, AFTER


middlewareManager = MiddlewareManager()



@middlewareManager.add_download_middleware(isBefore=True)
async def test_middleware(request=None, response=None, spider=None, item=None):
    print("test")
    print("test_middleware")





# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:50'

from typing import Callable, Dict
from sprite.http.headers import Headers


class Request:
    __slots__ = (
        '_encoding', '_query', 'method', '_formdata', 'priority', 'callback', 'cookies', '_headers', 'dont_filter',
        '_meta',
        '_url'
    )

    def __init__(self, url, callback: Callable = None, method='GET', headers=None, query=None, formdata=None,
                 cookies=None, meta=None, encoding='utf-8', priority=0,
                 dont_filter=False):
        self._encoding = encoding
        self._query = query
        self.method = str(method).upper()
        self._set_url(url)
        self._formdata = formdata
        assert isinstance(priority, int), "Request priority not an integer: %r" % priority
        self.priority = priority

        if callback is not None and not callable(callback):
            raise TypeError('callback must be a callable, got %s' % type(callback).__name__)

        self.callback = callback
        assert callback is not None, 'Request[%s]的回调函数[callback]不能为None!' % self._url
        self.cookies = cookies or {}
        self._headers = Headers(headers or {}, encoding=encoding)

        self.dont_filter = dont_filter

        self._meta = dict(meta) if meta else None

    @property
    def query(self):
        if self._query is None:
            return {}
        return self._query

    @property
    def formdata(self):
        if self._formdata is None:
            self._formdata = {}
        return self._formdata

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, headers:Dict):
        self._headers = Headers(headers, encoding=self._encoding)


    @property
    def meta(self):
        if self._meta is None:
            self._meta = {}
        return self._meta

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url: str):
        self._set_url(url)

    def _set_url(self, url):
        self._url = url

        if ':' not in self._url:
            raise ValueError('Missing scheme in request url: %s' % self._url)

    @property
    def encoding(self):
        return self._encoding

    def toUniqueStr(self):
        if self.method == "GET":
            return f'{self.url}'
        return f'{self.url},{self.formdata}'

    def __str__(self):
        return "<method:%s url:%s formdata:%s headers:%s cookies:%s>" % (
            self.method, self._url, self.formdata, self.headers.to_string(), self.cookies)

    __repr__ = __str__

    def copy(self):
        return self.replace()

    def replace(self, *args, **kwargs):
        """
        传入相关属性，生成一个新的request，不传入的属性，默认使用自身的属性
        """
        for x in ['url', 'method', 'headers', 'cookies', 'meta',
                  'encoding', 'priority', 'dont_filter', 'callback']:
            kwargs.setdefault(x, getattr(self, x))
        cls = kwargs.pop('cls', self.__class__)
        return cls(*args, **kwargs)




# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-24 19:27'

import sprite
from sprite import Spider, Response, Settings, Crawler
from examples.test_spider_one.middlewares import test_middleware
from examples.test_spider_one.items import TestItem


class GaodeSpider(Spider):
    name = "test_spider"

    async def start_request(self):
        headers = {
            'Accept': 'application/json, text/javascript, */*; q=0.01',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'zh-CN,zh;q=0.9',
            'Connection': 'keep-alive',
            'User-Agent': "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.86 Safari/537.36",
            'X-Requested-With': 'XMLHttpRequest',
        }
        start_requests = ["https://www.bilibili.com/index/recommend.json"]
        for url in start_requests:
            yield sprite.Request(url=url, headers=headers, callback=self.parse, dont_filter=True)

    async def parse(self, response: Response):
        self.logger.info("执行parse")
        self.logger.info(f'response body {response.body}')

        headers = {
            'Accept': 'application/json, text/javascript, */*; q=0.01',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'zh-CN,zh;q=0.9',
            'Connection': 'keep-alive',
            'User-Agent': "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.86 Safari/537.36",
            'X-Requested-With': 'XMLHttpRequest',
        }
        start_requests = ["https://apmconfig.douyucdn.cn/big/apm/front/config/report?client_sys=web",
                          "https://www.douyu.com/japi/search/api/getHotList"]
        for url in start_requests:
            self.logger.info("抛出一个请求！！！")
            yield sprite.Request(url=url, headers=headers, callback=self.test_parse, dont_filter=True)

    async def test_parse(self, response: Response):
        self.logger.info("test_parse")
        item = TestItem()
        item["body"] = response.body
        yield item


if __name__ == '__main__':
    # 实例化spider
    spider = GaodeSpider()
    # 实例化一个自定义的settings对象
    settings = Settings(values={
        "MAX_DOWNLOAD_NUM": 1,
        "WORKER_NUM": 1,
        "DELAY": 20,
        "LOG_FILE_PATH": "test_spider_one.log",
        "JOB_DIR": "/home/username/sprite/examples/test_spider_one/spiders",
        "LONG_SAVE": True,
    })
    # 构造一个crawler对象
    crawler = Crawler(
        spider=spider, middlewareManager=test_middleware, settings=settings)
    # 启动crawler
    crawler.run()



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-20 22:05'

import asyncio
from asyncio import AbstractEventLoop, wait_for

from collections import deque
from .connection import Connection, SECURE_CONTEXT, INSECURE_CONTEXT
from .proxy import ProxyHandler
from .exceptions import OpenConnectionFailed


class ConnectionPool:
    __slots__ = ('loop', 'host', 'port', 'protocol', 'max_connections', 'connections', 'available_connections',
                 'keep_alive', 'wait_connection_available', 'proxy', 'timeout')

    def __init__(self, loop: AbstractEventLoop, host: str, port: int, protocol: str, keep_alive: bool = True,
                 proxy: str = None):
        self.loop = loop
        self.host = host  # ip
        self.port = port  # 端口
        self.protocol = protocol  # 应用层协议 http/https
        self.available_connections = deque()  # 存储可用的闲置连接
        self.connections = set()  # 存储已经连接的链接
        self.keep_alive = keep_alive
        self.proxy = proxy
        # 代理模式下的链接不保存
        if self.proxy is not None:
            self.keep_alive = False

    async def create_connection(self, ssl=None) -> Connection:
        # 利用协程创建一个socket链接的参数
        args = {
            'host': self.host,
            'port': self.port,
            'loop': self.loop,
        }
        if self.proxy:
            self.proxy = ProxyHandler.prepend_scheme_if_needed(self.proxy, "http")
            proxy_protocol, proxy_host, proxy_port = ProxyHandler.parse_proxy(self.proxy)
            args = {
                'host': proxy_host,
                'port': proxy_port,
                'loop': self.loop,
            }
        # 配置https的ssl
        if self.proxy is None and self.protocol == 'https':
            if ssl is False:
                args['ssl'] = INSECURE_CONTEXT
            elif ssl is None:
                args['ssl'] = SECURE_CONTEXT
            else:
                args['ssl'] = ssl
        # 在协程里面创建一个socket连接
        #  (host=None, port=None, *, loop=None, limit=None, ssl=None, family=0, proto=0, flags=0, sock=None, local_addr=None, server_hostname=None, ssl_handshake_timeout=None)
        # 会一直阻塞着等待连接
        reader, writer = await asyncio.open_connection(**args)
        # 实例化一个connection
        connection = Connection(self.loop, reader, writer, self)
        self.connections.add(connection)
        return connection

    async def get_connection(self, ssl):
        try:
            # 从可用链接队列尾部里面取出一个链接
            connection = self.available_connections.pop()
            if not await connection.is_dropped():
                # 从可用链接中获取到一个connection，且能正常读取
                return connection
            else:
                # 可用链接读取超时
                await self.release_connection(connection)
                # 创建新的链接
                return await self.create_connection(ssl)
        except IndexError:
            # 可用链接队列为空，无法pop
            return await self.create_connection(ssl)

    async def release_connection(self, connection: Connection, keep_alive: bool = True):
        if keep_alive and self.keep_alive:
            # 单个链接是长链接，且域名链接是长链接，则把链接放在可用链接队列的头部
            self.available_connections.appendleft(connection)
        else:
            # 只要有一个设置的不是长链接
            # 关闭链接
            connection.close()  # 关闭链接需要等待
            self.connections.discard(connection)

    def close(self):
        # 关闭域名下的所有链接
        for connection in self.connections:
            connection.close()



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-17 22:00'

import time
from typing import Dict, Callable
from collections import deque
from sprite.http.request import Request
from sprite.exceptions import TypeNotSupport


def request_to_dict(request: Request) -> Dict:
    if not isinstance(request, Request):
        raise TypeNotSupport("不能转换类型非Request的实例为字典！")
    d = {
        'url': request.url,
        'callback': request.callback.__name__,
        'method': request.method,
        'headers': dict(request.headers),
        'cookies': request.cookies,
        'meta': request.meta,
        '_encoding': request._encoding,
        'priority': request.priority,
        'dont_filter': request.dont_filter,
    }

    return d


def request_from_dict(spider:"Spider", d: Dict) -> Request:
    if not isinstance(d, Dict):
        raise TypeNotSupport("不能转换类型非字典对象为Request实例")
    return Request(
        url=d["url"],
        callback=_get_method(spider, d["callback"]),
        method=d['method'],
        headers=d['headers'],
        cookies=d['cookies'],
        meta=d['meta'],
        encoding=d['_encoding'],
        priority=d['priority'],
        dont_filter=d['dont_filter'])

def _get_method(obj, method_name: str) -> Callable:
    try:
        return getattr(obj, method_name)
    except AttributeError:
        raise ValueError(f'method {method_name} not found in {obj}')


class Counter:
    def __init__(self, unit: int = 60):
        self._first_time: float = time.time()
        self._first_count = 0
        self._travel_record = deque()
        self._count = 0
        self._unit = unit

    def dot(self) -> [int, int]:
        speed = None
        unit_count = None

        now = time.time()
        passed = now - self._first_time
        self._count += 1
        if passed < self._unit:
            self._travel_record.append((now, self._count))
        else:
            speed, unit_count = self._getUintSpedd(), self._getUintCount()
            # 更新单位时间点的数据
            self._first_time, self._first_count = now - \
                (now - self._unit), self._count
            # 清空上个单位时间段的记录
            self._travel_record.clear()
            # 记录这个时间段的第一笔数据
            self._travel_record.append((now, self._count))
        return speed, unit_count

    def _getUintSpedd(self) -> float:
        try:
            _, last_count = self._travel_record[-1]
            return (last_count - self._first_count) / self._unit
        except IndexError:
            return 0

    def _getUintCount(self) -> int:
        try:
            _, last_count = self._travel_record[-1]
            return last_count - self._first_count
        except IndexError:
            return 0

    @property
    def unit(self) -> int:
        return self._unit



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:50'

from collections import Mapping


class CaselessDict(dict):
    __slots__ = ()

    def __init__(self, seq=None):
        super(CaselessDict, self).__init__()
        if seq:
            self.update(seq)

    def __getitem__(self, key):
        return dict.__getitem__(self, self.normkey(key))

    def __setitem__(self, key, value):
        dict.__setitem__(self, self.normkey(key), self.normvalue(value))

    def __delitem__(self, key):
        dict.__delitem__(self, self.normkey(key))

    def __contains__(self, key):
        return dict.__contains__(self, self.normkey(key))

    has_key = __contains__

    def __copy__(self):
        return self.__class__(self)

    copy = __copy__

    def normkey(self, key):
        """Method to normalize dictionary key access"""
        try:
            return key.lower()
        except TypeError:
            tmp_str = ""
            for c in key:
                try:
                    tmp_str += c.lower()
                except TypeError:
                    tmp_str += c
        return tmp_str

    def normvalue(self, value):
        """Method to normalize values prior to be setted"""
        return value

    def get(self, key, def_val=None):
        return dict.get(self, self.normkey(key), self.normvalue(def_val))

    def setdefault(self, key, def_val=None):
        return dict.setdefault(self, self.normkey(key), self.normvalue(def_val))

    def update(self, seq):
        seq = seq.items() if isinstance(seq, Mapping) else seq
        iseq = ((self.normkey(k), self.normvalue(v)) for k, v in seq)
        super(CaselessDict, self).update(iseq)

    @classmethod
    def fromkeys(cls, keys, value=None):
        return cls((k, value) for k in keys)

    def pop(self, key, *args):
        return dict.pop(self, self.normkey(key), *args)






# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-25 13:14'

from .headers import Headers
from .response import Response

HTTP_PARSE_ENCODING = 'utf-8'

"""
http response parser
"""

"""HttpRequestParser

        protocol -- a Python object with the following methods
        (all optional):

          - on_message_begin()
          - on_url(url: bytes)
          - on_header(name: bytes, value: bytes)
          - on_headers_complete()
          - on_body(body: bytes)
          - on_message_complete()
          - on_chunk_header()
          - on_chunk_complete()
          - on_status(status: bytes)
"""


class HttpToolsProtocol:
    def __init__(self, response:Response,  expected_content_length: int = 1024 * 1024):
        self._response= response
        self._status = 0
        self._content = b''
        self._headers = Headers()
        # Response state
        self._response_started = False
        self._response_complete = False
        self._headers_complete = False
        self._chunked_encoding = None
        self._expected_content_length = expected_content_length

    def on_message_begin(self):
        self._response_started = True

    def on_url(self, url: bytes):
        pass

    def on_header(self, name: bytes, value: bytes):
        self._headers[name.decode(HTTP_PARSE_ENCODING)] = value.decode(HTTP_PARSE_ENCODING)

    def on_headers_complete(self):
        self._headers_complete = True
        self._response.headers = self._headers

    def on_body(self, body: bytes):
        self._content += body

    def on_message_complete(self):
        self._response_complete = True

    def on_chunk_header(self):
        self._chunked_encoding = True

    def on_chunk_complete(self):
        self._response_complete = True

    def on_status(self, status: bytes):
        self._status = status.decode(HTTP_PARSE_ENCODING)

    @property
    def headers(self):
        return self._headers

    @property
    def content(self):
        return self._content

    def __str__(self):
        return f'{self._status}   {self._headers}   {self._content}'



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/26 20:19'

import base64
from typing import Dict
from urllib.parse import urlparse, unquote, urlunparse


class ProxyHandler:

    @classmethod
    def get_auth_from_url(cls, url: str):
        parsed = urlparse(url)

        try:
            auth = (unquote(parsed.username), unquote(parsed.password))
        except (AttributeError, TypeError):
            auth = ('', '')

        return auth

    @classmethod
    def get_proxy_headers(cls, proxy: str):
        headers = {}
        username, password = cls.get_auth_from_url(proxy)

        if username:
            headers['Proxy-Authorization'] = cls._basic_auth_str(username,
                                                                 password)

        return headers

    @classmethod
    def _basic_auth_str(cls, username: str, password: str):
        if not isinstance(username, str):
            username = str(username)

        if not isinstance(password, str):
            password = str(password)

        if isinstance(username, str):
            username = username.encode('latin1')

        if isinstance(password, str):
            password = password.encode('latin1')

        proxy_user_pass = f'{username}:{password}'
        authstr = "Basic " + base64.urlsafe_b64encode(bytes((proxy_user_pass), "ascii")).decode("utf8")
        return authstr

    @classmethod
    def select_proxy(cls, url: str, proxies: Dict):
        proxies = proxies or {}
        urlparts = urlparse(url)
        if urlparts.hostname is None:
            return proxies.get(urlparts.scheme, proxies.get('all'))

        proxy_keys = [
            urlparts.scheme + '://' + urlparts.hostname,
            urlparts.scheme,
            'all://' + urlparts.hostname,
            'all',
        ]
        proxy = None
        for proxy_key in proxy_keys:
            if proxy_key in proxies:
                proxy = proxies[proxy_key]
                break

        return proxy

    @classmethod
    def prepend_scheme_if_needed(cls, url: str, new_scheme: str):

        scheme, netloc, path, params, query, fragment = urlparse(url)
        if scheme != new_scheme:
            scheme = new_scheme
        return urlunparse((scheme, netloc, path, params, query, fragment))

    @classmethod
    def parse_proxy(cls, url: str):
        parse_result = urlparse(url)
        return parse_result.scheme, parse_result.netloc.split(":")[0], parse_result.port



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-25 23:41'



class Headers:
    __slots__ = ('raw', 'values', 'evaluated')

    def __init__(self, raw=None):
        self.raw = raw or []
        self.values = None
        self.evaluated = False

    def get(self, key, default=None):
        if not self.evaluated:
            self.eval()
        return self.values.get(key.lower()) or default

    def eval(self):
        self.values = {}
        while self.raw:
            header = self.raw.pop()
            self.values[header[0].decode('utf-8').lower()] = header[1].decode('utf-8')
        self.evaluated = True

    def dump(self):
        if not self.evaluated:
            self.eval()
        return self.values

    def parse_cookies(self) -> dict:
        header = self.get('cookie')
        cookies = {}
        if header:
            for cookie in header.split(';'):
                first = cookie.find('=')
                name = cookie[:first].strip()
                value = cookie[first + 1:]
                cookies[name] = value
        return cookies

    def __getitem__(self, item: str):
        if not self.evaluated:
            self.eval()
        return self.values[item.lower()]

    def __setitem__(self, key: str, value: str):
        if not self.evaluated:
            self.eval()
        self.values[key.lower()] = value

    def __repr__(self):
        return f'<Headers {self.dump()}>'


# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-21 00:43'

import asyncio
# from queue import Queue
import uuid
from typing import Union
from urllib.parse import urlencode
from typing import List
from asyncio import AbstractEventLoop, TimeoutError
from .limits import RequestRate
from .pool import ConnectionPool
from .url import URL
from .request import Request
from .response import Response
from sprite.http.cookies import CookiesJar
from sprite.exceptions import TooManyRedirects, MissingSchema
from .url import parse_url
from sprite.utils.decoders import MultipartEncoder
from sprite.http.cookies import SessionCookiesJar
import json as json_module
from .retries import RetryStrategy

URL_ENCODING = 'utf-8'
BODY_ENCODING = 'utf-8'


class ClientDefaults:
    TIMEOUT = 30
    HEADERS = {
        'User-Agent': 'Vibora',
        'Accept': '*/*',
        'Connection': 'keep-alive',
        'Accept-Encoding': 'gzip, deflate'
    }
    RETRY_STRATEGY = RetryStrategy()


class HTTPEngine:
    __slots__ = ('loop', 'session', 'pools', 'limits')

    def __init__(self, session: 'Session', loop: AbstractEventLoop, limits: List[RequestRate] = None):
        self.loop = loop
        self.session = session  # 对外的调用端入口实例
        self.pools = {}  # 多个connection_pool链接管理，每一个connection_pool对应一个地址和端口下的多个链接
        self.limits = limits or []

    def get_pool(self, protocol: str, host: str, port: str, proxy: str = None) -> ConnectionPool:
        key = (protocol, host, port)
        if port in (0, None):
            if protocol == "https":
                port = 443
            else:
                port = 80
        try:
            # 尝试从字典中获取指定域名的链接池
            return self.pools[key]
        except KeyError:
            self.pools[key] = ConnectionPool(loop=self.loop, host=host, port=port, protocol=protocol,
                                             keep_alive=self.session.keep_alive, proxy=proxy)
        return self.pools[key]

    async def handle_redirect(self, request: Request, response: Response, stream: bool, follow_redirects: bool,
                              max_redirects: int, decode: bool, validate_ssl, headers: dict,
                              response_cookies: CookiesJar) -> Response:
        if max_redirects == 0:
            # 禁止跳转
            raise TooManyRedirects
        try:
            location = response.headers["location"]
        except KeyError:
            raise Exception('HTTP redirect response missing location header.')
        if not location.startswith('http'):
            if not location.startswith('/'):
                location = '/' + location
            location = request.url.netloc + location
        # 解析跳转的url,返回URL实例
        redirect_url = parse_url(location.encode())
        headers['Host'] = redirect_url.host
        # 再次对跳转的URL发起请求
        return await self.request(
            url=redirect_url, method='GET', stream=stream, follow_redirects=follow_redirects,
            max_redirects=(max_redirects - 1), decode=decode, validate_ssl=validate_ssl, headers=headers,
            origin=response, cookies=response_cookies
        )

    async def throttle(self, url: str):
        # 对 需要打卡的URL进行记录
        for limit in self.limits:
            if not limit.pattern or limit.pattern.fullmatch(url):
                # 目前的limit记录的URL没有匹配模式，或者传入的URL通过了匹配模式
                await limit.notify()

    async def request(self, url: URL, method: str, stream: bool, follow_redirects: bool,
                      max_redirects: int, decode: bool, validate_ssl, headers: dict,
                      origin: Response = None, data=None, cookies: CookiesJar = None, proxy: str = None) -> Response:
        """
        先建立链接，再发起请求，分为两个阶段
        """

        # 发起链接，对指定的URL，limit限制的是发起链接的频率而不是
        if self.limits:
            await self.throttle(url.raw)
        # 1.发起一个链接池链接
        # 设置请求的域名
        headers["Host"] = url.host
        pool = self.get_pool(url.schema, url.host, url.port, proxy=proxy)
        connection = await pool.get_connection(validate_ssl)
        # 从连接引擎对象上存储的cookies容器来拉取已经请求过对应的域名，保留下来的cookies

        # 2. 实例话一个request对象，来封装这一次请求
        if cookies is None:
            cookies = self.session.cookies.get(domain=url.host)
        request = Request(method, url, headers, data, cookies=cookies,
                          origin=origin, )
        # 编码请求信息，并发起请求
        await request.encode(connection)

        # 3.实例话一个response对象来管理接受响应
        response = Response(request.url, connection,
                            request=request, decode=decode)
        # 4.接受headers
        await response.receive_headers()
        # 5.接受完毕headers之后，再把headers里面可能的cookie合并存储到客户端全局的cookies容器中
        response_cookies = await response.cookies
        self.session.cookies.merge(response_cookies, domain=request.url.host)
        # 6.通过读取的headers里面的信息，和客户端的配置项来决定是否跳转
        if follow_redirects:
            if response.is_redirect():
                await response.read_content()
                return await self.handle_redirect(request, response, stream, follow_redirects,
                                                  max_redirects, decode, validate_ssl, headers, response_cookies)
        if not stream:
            # 7.如果不是流式数据，直接读取指定长度的数据即可
            await response.read_content()
        # 返回响应对象实例
        return response

    def close(self):
        # 关闭所有的链接池
        for pool in self.pools.values():
            pool.close()


class Session:
    __slots__ = ('_loop', '_downloader', '_engine', '_headers', 'follow_redirects', 'max_redirects',
                 'stream', 'decode', 'ssl', 'prefix', 'keep_alive', 'retries_policy',
                 'timeout', 'cookies', 'limits')

    def __init__(self, loop: AbstractEventLoop = None, downloader: 'Downloader' = None, headers: dict = None,
                 follow_redirects: bool = False, max_redirects: int = 30,
                 stream: bool = False, decode: bin = True, ssl=None, keep_alive: bool = True,
                 prefix: str = '', timeout: Union[int, float] = ClientDefaults.TIMEOUT,
                 retries: RetryStrategy = None, limits: List[RequestRate] = None):
        self._downloader = downloader
        self._loop = loop or asyncio.get_event_loop()
        # 全局的connection链接管理
        self._engine = HTTPEngine(self, self._loop, limits=limits)
        # 全局的headers
        self._headers = ClientDefaults.HEADERS
        if headers:
            self._headers.update(headers)
        self.follow_redirects = follow_redirects
        self.max_redirects = max_redirects
        self.stream = stream
        self.decode = decode
        self.ssl = ssl
        self.prefix = prefix.encode(URL_ENCODING) or b''  # 设定URL饿前缀域名
        self.keep_alive = keep_alive
        self.retries_policy = retries or ClientDefaults.RETRY_STRATEGY
        self.timeout = timeout
        # 全局的session管理
        self.cookies = SessionCookiesJar()  # 会话管理对象的cookie默认是无法设置，初始状态为空
        self.limits = limits

    @staticmethod
    def build_url(prefix: bytes, url: bytes, query: dict):
        if not url:
            raise ValueError('Url parameter must not be empty.')
        if prefix and prefix.endswith(b'/'):
            if url.startswith(b'/'):
                url = url[1:]
            url = prefix + url
        elif prefix:
            if url.startswith(b'/'):
                url = prefix + url
            else:
                url = prefix + b'/' + url

        if not url.startswith(b'http'):
            raise MissingSchema(f'Missing schema in {url.decode(URL_ENCODING)}. '
                                f'Perhaps you meant http://{url.decode(URL_ENCODING)} ?.')
        if query:
            url = url + b'?' + urlencode(query).encode(URL_ENCODING)

        return url

    async def request(self, url: str = '/', stream: bool = None, follow_redirects: bool = None,
                      max_redirects: int = 30, decode: bool = None, ssl=None, timeout: int = None,
                      retries: Union[RetryStrategy, int] = None, cookies: dict = None,
                      headers: dict = None, method: str = 'GET', query: dict = None,
                      json: dict = None, ignore_prefix: bool = False, body=None,
                      form: dict = None, proxy: str = None) -> Response:

        # Asserting the user is not using conflicting params.
        if sum([body is not None, json is not None, form is not None]) > 1:
            raise ValueError(
                'You cannot set body, json or form together. You must pick one and only one.')

        # Handling default parameters.
        stream = stream if stream is not None else self.stream
        follow_redirects = follow_redirects if follow_redirects is not None else self.follow_redirects
        max_redirects = max_redirects if max_redirects is not None else self.max_redirects
        decode = decode if decode else self.decode
        timeout = timeout if timeout else self.timeout
        ssl = ssl if ssl is not None else self.ssl
        # 配置重试策略
        # retries = retries.clone() if retries is not None else RetryStrategy()

        request_headers = self._headers.copy()
        if cookies is not None:
            cookies = CookiesJar(cookies)
        if headers:
            # 用自定义的headers更新默认的headers
            request_headers.update(headers)

        # Constructing the URL.
        # 利用传入的URL和客户端保存的URL前缀来拼接出完整的URL，是一个字节数组
        url = self.build_url(prefix=self.prefix if not ignore_prefix else b'', url=url.encode(URL_ENCODING),
                             query=query)
        # 解析字节数组成字符串URL
        parsed_url = parse_url(url)

        if json is not None:
            body = json_module.dumps(json).encode('utf-8')
            request_headers['Content-Type'] = 'application/json'

        if form is not None:
            boundary = str(uuid.uuid4()).replace('-', '').encode()
            body = MultipartEncoder(delimiter=boundary, params=form)
            request_headers['Content-Type'] = f'multipart/form-data; boundary={boundary.decode()}'

        while True:
            try:
                task = self._engine.request(
                    url=parsed_url, data=body, method=method, stream=stream,
                    follow_redirects=follow_redirects, max_redirects=max_redirects, decode=decode,
                    validate_ssl=ssl, headers=request_headers, cookies=cookies, proxy=proxy)
                if timeout:
                    response = await asyncio.wait_for(task, timeout)
                else:
                    response = await task
                return response
            except (ConnectionError, TimeoutError) as error:
                raise error

            # 这里触发多次尝试的机制，失败则抛出error
            # try:
            #     task = self._engine.request(
            #         url=parsed_url, data=body, method=method, stream=stream,
            #         follow_redirects=follow_redirects, max_redirects=max_redirects, decode=decode,
            #         validate_ssl=ssl, headers=request_headers, cookies=cookies)
            #     if timeout:
            #         response = await asyncio.wait_for(task, timeout)
            #     else:
            #         response = await task
            #     if retries.responses.get(response.status_code, 0) > 0:
            #         retries.responses[response.status_code] -= 1
            #         continue
            #     return response
            # except (ConnectionError, TimeoutError) as error:
            #     if retries.network_failures.get(method, 0) > 0:
            #         retries.network_failures[method] -= 1
            #         continue
            #     raise error

    # session的入口方法， 这个方法在超时或者连接失败或者多次尝试失败的情况下，会抛出异常的error（ConnectionError, TimeoutError）
    async def get(self, url: str = '', stream: bool = None, follow_redirects: bool = None, max_redirects: int = 30,
                  decode: bool = None, ssl=None, timeout: int = None,
                  retries: RetryStrategy = None, headers: dict = None, cookies: dict = None,
                  query: dict = None, ignore_prefix: bool = False, proxy: str = None) -> Response:

        response = await self.request(url=url, stream=stream, follow_redirects=follow_redirects,
                                      max_redirects=max_redirects, decode=decode, ssl=ssl, cookies=cookies,
                                      retries=retries, headers=headers, timeout=timeout, method='GET', query=query,
                                      ignore_prefix=ignore_prefix, proxy=proxy)
        return response

    # session的入口方法， 这个方法在超时或者连接失败或者多次尝试失败的情况下，会抛出异常的error（ConnectionError, TimeoutError）
    async def post(self, url: str = '', stream: bool = None, follow_redirects: bool = None, max_redirects: int = 30,
                   decode: bool = None, ssl=None, timeout=ClientDefaults.TIMEOUT, cookies: dict = None,
                   retries: Union[RetryStrategy, int] = None, headers: dict = None, query: dict = None,
                   body=None, form=None, json=None, ignore_prefix: bool = False, proxy: str = None) -> Response:
        response = await self.request(url=url, stream=stream, follow_redirects=follow_redirects,
                                      max_redirects=max_redirects, decode=decode, ssl=ssl, cookies=cookies,
                                      retries=retries, headers=headers, timeout=timeout, method='POST', query=query,
                                      ignore_prefix=ignore_prefix, body=body, form=form, json=json, proxy=proxy)
        return response

    def close(self):
        """

        :return:
        """
        self._engine.close()

    async def __aenter__(self):
        """

        :return:
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.close()
        await asyncio.sleep(0)
        if exc_val:
            raise exc_val



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-24 19:09'

from sprite.middlewaremanager import MiddlewareManager

test_middleware = MiddlewareManager()


# 增加下载中间件，在下载request之前执行
@test_middleware.add_download_middleware(isBefore=True)
async def test_downloader_middleware(request, spider=None):
    spider.logger.info("这是下载中间件  before")
    headers = {
        'Accept': 'application/json, text/javascript, */*; q=0.01',
        'Accept-Encoding': 'gzip, deflate',
        'Accept-Language': 'zh-CN,zh;q=0.9',
        'Connection': 'keep-alive',
        'User-Agent': "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.86 Safari/537.36",
        'X-Requested-With': 'XMLHttpRequest',
    }
    request.headers = headers
    spider.logger.info(f'捕获request   {request.url}')


# 增加下载中间件，在下载request之后执行
@test_middleware.add_download_middleware(isBefore=False)
async def test_downloader_middleware(response=None, spider=None):
    spider.logger.info("这是下载中间件 after")
    spider.logger.info(f'捕获response   {response.status}')


# 增加处理item中间件，获取item之后执行
@test_middleware.add_pipeline_middleware()
async def test_downloader_middleware(item=None, spider=None):
    spider.logger.info("这是管道中间件")
    spider.logger.info(f'捕获item   {item}')


# 增加spider中间件，在启动spider之前执行
@test_middleware.add_spider_middleware(isBefore=True)
async def test_downloader_middleware(spider=None):
    spider.logger.info("这是spider中间件 before")


# 增加spider中间件，在启动spider之后执行
@test_middleware.add_spider_middleware(isBefore=False)
async def test_downloader_middleware(spider=None):
    spider.logger.info("这是spdier中间件 after")



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/31 16:37'

from typing import Callable, Coroutine
from types import AsyncGeneratorType, GeneratorType


async def detailCallable(callable_obj, result_detail_func: Callable):
    call_result = callable_obj()
    if isinstance(call_result, AsyncGeneratorType):
        async for callback_result in call_result:
            detail_result = result_detail_func(callback_result)
            if isinstance(detail_result, Coroutine):
                await detail_result
    elif isinstance(call_result, Coroutine):
        callback_result = await call_result
        result = result_detail_func(callback_result)
        if isinstance(result, Coroutine):
            await result

    elif isinstance(call_result, GeneratorType):
        for callback_result in call_result:
            detail_result = result_detail_func(callback_result)
            if isinstance(detail_result, Coroutine):
                await detail_result
    else:
        detail_result = result_detail_func(call_result)
        if isinstance(detail_result, Coroutine):
            await detail_result



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:43'

from .http.request import Request
from .http.response import Response
from .item import Item
from .settings import Settings
from .spider import Spider
from .crawl import Crawler
from .core.download import Downloader
from .item import Field
from .utils.coroutinePool import PyCoroutinePool


__version__ = "0.1.0"



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-20 00:50'

import asyncio
import traceback
from queue import Queue
from typing import Union, List
from asyncio import AbstractEventLoop, TimeoutError
from .limits import RequestRate
from .pool import ConnectionPool
from sprite.http.request import Request
from sprite.http.response import Response
from .retries import RetryStrategy
from sprite.utils.coroutinePool import PyCoroutinePool
from .session import ClientDefaults, Session
from sprite.settings import Settings
from sprite.utils.log import get_logger
from .limits import RequestRate

logger = get_logger()


class Downloader:
    __slots__ = (
        '_session', '_loop', '_coroutine_pool', '_downloaded_response', '_downloaded_request', '_crawler', '_settings',
        '_sem', '_delay', '_no_complete_task')

    def __init__(self, loop: AbstractEventLoop, max_download_num: int = 5, coroutine_pool: PyCoroutinePool = None,
                 headers: dict = None,
                 follow_redirects: bool = False, max_redirects: int = 30, delay: int = 1,
                 stream: bool = False, decode: bin = True, ssl=None, keep_alive: bool = True,
                 prefix: str = '', timeout: Union[int, float] = ClientDefaults.TIMEOUT,
                 retries: RetryStrategy = None, limits: List[RequestRate] = None):
        self._loop = loop
        self._sem = asyncio.Semaphore(max_download_num)
        self._delay = delay
        self._no_complete_task = 0

        self._session = Session(loop=self._loop, headers=headers, follow_redirects=follow_redirects,
                                max_redirects=max_redirects,
                                stream=stream, decode=decode, ssl=ssl, keep_alive=keep_alive, prefix=prefix,
                                timeout=timeout,
                                retries=retries, limits=limits)
        # 协程池
        if coroutine_pool is None:
            self._coroutine_pool = PyCoroutinePool()
            self._coroutine_pool.start()
        # 自带的缓冲队列
        self._downloaded_response = Queue()

    # 添加下载任务
    def addTask(self, request: Request):
        task = self.request(request)
        # 添加到协程池中执行
        self._coroutine_pool.go(task=task)
        self._no_complete_task += 1

    def getResponse(self) -> Union[Response, Exception]:
        # 阻塞获取response
        response = self._downloaded_response.get()
        return response

    def addResponse(self, response: Union[Response, Exception]):
        self._downloaded_response.put(response)


    def _get_proxy(self, request: Request):
        return request.meta.get("proxy", None)

    async def request(self, request: Request) -> Union[Response, Exception]:
        # 强制每一个请求的下载间隔一小会
        if self._delay > 0:
            await asyncio.sleep(self._delay)
        # 加锁的目的是为了限制同一时间的并发数量
        async with self._sem:
            try:
                proxy = self._get_proxy(request)
                if request.method == "GET":
                    response = await self._session.get(url=request.url, headers=request.headers,
                                                       query=request.query,
                                                       cookies=request.cookies, ignore_prefix=True, proxy=proxy)
                else:
                    response = await self._session.post(url=request.url, headers=request.headers,
                                                        query=request.query,
                                                        form=request.formdata, cookies=request.cookies,
                                                        ignore_prefix=True, proxy=proxy)
                response = Response(url=response.url, status=response.status_code, headers=response.headers.dump(),
                                    body=response.content.decode("utf-8"), request=request, )
                self.addResponse(response)
                return response
            # 请求失败
            except (ConnectionError, TimeoutError) as error:
                if isinstance(error, ConnectionError):
                    logger.info(f'find one connection of error')
                elif isinstance(error, TimeoutError):
                    logger.info(f'find one timeout of error')
                response = Response(url=request.url, status=400, headers=request.headers,
                                    request=request, error=error)
                self.addResponse(response)
                return response
            except Exception as error:
                logger.info(f'下载过程中发现一个错误！丢失request！\n{traceback.format_exc()}')
                self.addResponse(error)
                raise error

            finally:
                self._no_complete_task -= 1


    def close(self):
        self._session.close()

    def has_pending_response(self):
        if self._downloaded_response.empty():
            return False
        else:
            return True

    def has_no_complete_task(self) -> bool:
        if self._no_complete_task > 0:
            return True
        return False

    @classmethod
    def from_settings(cls, settings: Settings, coroutine_pool: PyCoroutinePool):
        max_download_num = settings.getint("MAX_DOWNLOAD_NUM")
        headers = settings.getdict("HEADERS")
        follow_redirects = settings.getbool("FOLLOW_REDIRECTS")
        max_redirects = settings.getint("MAX_REDIRECTS")
        stream = settings.getbool("STREAM")
        decode = settings.getbool("DECODE")
        ssl = settings.get("SSL")
        keep_alive = settings.getbool("KEEP_ALIVE")
        prefix = settings.get("PREFIX")
        timeout = settings.get("TIMEOUT")
        limits = settings.get("LIMITS")
        delay = settings.getint("DELAY")

        obj = cls(loop=coroutine_pool.loop, max_download_num=max_download_num, coroutine_pool=coroutine_pool,
                  headers=headers, follow_redirects=follow_redirects, max_redirects=max_redirects, delay=delay,
                  stream=stream, decode=decode, ssl=ssl, keep_alive=keep_alive, prefix=prefix,
                  timeout=timeout, limits=limits)
        return obj



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-17 20:50'

import logging

from sprite.settings import Settings


def get_logger(name: str = "sprite"):
    return logging.getLogger(name)



def set_logger(settings: Settings):
    file_path = settings.get("LOG_FILE_PATH", "")
    # project_name = settings.get("PROJECT_NAME")
    logging_format = "[%(name)s %(asctime)s %(levelname)s]: "
    # logging_format = "%(name)s [%(asctime)s %(levelname)s %(filename)s]: "
    # logging_format += "%(module)-7s::l%(lineno)d: "
    # logging_format += "%(module)-7s: "
    logging_format += "%(message)s"

    logging.basicConfig(
        filename=file_path,
        format=logging_format,
        level=logging.INFO,
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    logging.getLogger("asyncio").setLevel(logging.INFO)
    logging.getLogger("websockets").setLevel(logging.INFO)



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:45'

import os
import pickle
import traceback
from typing import Union
from sprite.utils.queues import Queue
from sprite.http.request import Request
from sprite.utils.queues import PriorityQueue
from sprite.utils.log import get_logger
from sprite.utils.pybloomfilter import ScalableBloomFilter
from sprite.utils.request import request_to_dict, request_from_dict
from sprite.settings import Settings
from sprite.exceptions import SchedulerEmptyException, TypeNotSupport, SchedulerEmptyException


logger = get_logger()


# 正在处理中的Request
class Slot:
    def __init__(self):
        self.process_request = Queue()

    def addRequest(self, req: Request):
        self.process_request.put_nowait(req)

    def getRequest(self) -> Request:
        return self.process_request.get_nowait()

    def has_pending_request(self) -> bool:
        return not self.process_request.empty()

    # 告诉队列，一个任务已经完成
    def toDone(self):
        self.process_request.task_done()

    async def join(self):
        await self.process_request.join()


# 调度器
class Scheduler:
    def __init__(self, spider:"Spider", df=None, queue=None, long_save: bool = False, job_dir: str = None):
        self._spider=spider
        self._priorityQueue = queue or PriorityQueue()
        self._df = df or ScalableBloomFilter()
        self._long_save = long_save
        self._job_dir = job_dir

    @classmethod
    def from_settings(cls, settings: Settings, spider:"Spider"):
        initial_capacity = settings.getint("INITIAL_CAPACITY")
        error_rate = settings.getfloat("ERROR_RATE")
        long_save = settings.getbool("LONG_SAVE")
        job_dir = settings.get("JOB_DIR")
        obj = cls(spider=spider, df=ScalableBloomFilter(initial_capacity=initial_capacity),
                  queue=PriorityQueue(), long_save=long_save, job_dir=job_dir)
        return obj

    def has_pending_requests(self):
        return len(self._priorityQueue) > 0

    # 请求加入队列
    def enqueue_request(self, request: Request) -> bool:
        request_unique = request.toUniqueStr()
        if not request.dont_filter and request_unique in self._df:
            return True
        self._df.add(request_unique)
        self._priorityQueue.push(request, request.priority)
        return True

    # 取出request
    def next_request(self) -> Request:
        try:
            request, _ = self._priorityQueue.pop()
        except IndexError:
            # 队列为空
            raise SchedulerEmptyException("scheduler is empty")
        return request

    def __len__(self):
        return len(self._priorityQueue)

    def close(self):
        try:
            self._save_requests()
        except:
            logger.info(
                f'close scheduler, find one error: \n{traceback.format_exc()}')

    def start(self):
        try:
            self._load_requests()
        except:
            logger.info(
                f'start scheduler, find one error: \n{traceback.format_exc()}')

    def _save_requests(self):
        if self._long_save and self._job_dir is not None and os.path.exists(self._job_dir) and self.has_pending_requests():
            logger.info(f'save request')
            requests = []
            while True:
                try:
                    request = self.next_request()
                except SchedulerEmptyException:
                    break
                try:
                    requests.append(request_to_dict(request))
                except TypeNotSupport:
                    logger.info(
                        f'find one error: \n{traceback.format_exc()}')

            requests_file_path = os.path.join(self._job_dir, "requests.pickle")
            with open(requests_file_path, "wb") as f:
                pickle.dump(requests, f)

    def _load_requests(self):
        if self._long_save and self._job_dir is not None:
            requests_file_path = os.path.join(self._job_dir, "requests.pickle")
            if os.path.exists(requests_file_path):
                logger.info(f'load request')
                with open(requests_file_path, "rb") as f:
                    requests = pickle.load(f)
                for request in requests:
                    try:
                        self.enqueue_request(request_from_dict(self._spider, request))
                    except TypeNotSupport:
                        logger.info(
                            f'find one error: \n{traceback.format_exc()}')
                # 删除requests缓存
                os.remove(requests_file_path)




# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:27'


class NotSupported(Exception): pass


class HTTPClientError(Exception): pass


class StreamAlreadyConsumed(HTTPClientError): pass


class TypeNotSupport(Exception): pass


class TooManyRedirects(HTTPClientError):pass


class MissingSchema(HTTPClientError):pass


class NotStartRequest(Exception): pass


class DownloadException(Exception): pass


class SchedulerEmptyException(Exception): pass


# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-24 18:46'


from sprite.settings import Settings





if __name__ == '__main__':
    class TestDict:
        def __init__(self):
            pass

    test_settings_dict = {
        "name":"liyong",
        "test":TestDict(),
    }

    settings = Settings(values=test_settings_dict)
    print(settings.get("test"))
    print(settings.getdict("HEADERS"))
    print(settings.get("MOSTSTOP"))
    print(settings.getint("MAXCOROUTINEAMOUNT"))










# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-19 23:28'

from typing import Union, List, Dict
import json
import copy
from collections import MutableMapping
from importlib import import_module
from pprint import pformat

from . import default_settings  # 用的相对路径，便于相对导入

SETTINGS_PRIORITIES = {
    'default': 0,
    'command': 10,
    'project': 20,
    'spider': 30,
    'cmdline': 40,
}


# 优先级等级名称与等级数值直接的转换
def get_settings_priority(priority: Union[str, int]) -> int:
    if isinstance(priority, str):
        return SETTINGS_PRIORITIES[priority]
    else:
        return priority


class SettingsAttribute(object):

    def __init__(self, value, priority):
        self.value = value
        if isinstance(self.value, BaseSettings):
            self.priority = max(self.value.maxpriority(), priority)
        else:
            self.priority = priority

    def set(self, value, priority):
        """Sets value if priority is higher or equal than current priority."""
        if priority >= self.priority:
            if isinstance(self.value, BaseSettings):
                value = BaseSettings(value, priority=priority)
            self.value = value
            self.priority = priority

    def __str__(self):
        return "<SettingsAttribute value={self.value!r} " \
               "priority={self.priority}>".format(self=self)

    __repr__ = __str__


class BaseSettings(MutableMapping):
    def __init__(self, values=None, priority='project'):
        self.frozen = False
        self.attributes = {}
        self.update(values, priority)

    def __getitem__(self, opt_name):
        if opt_name not in self:
            return None
        return self.attributes[opt_name].value

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name, default=None):

        return self[name] if self[name] is not None else default

    def getbool(self, name, default=False):

        got = self.get(name, default)
        try:
            return bool(int(got))
        except ValueError:
            if got in ("True", "true"):
                return True
            if got in ("False", "false"):
                return False
            raise ValueError("Supported values for boolean settings "
                             "are 0/1, True/False, '0'/'1', "
                             "'True'/'False' and 'true'/'false'")

    def getint(self, name, default=0):
        return int(self.get(name, default))

    def getfloat(self, name, default=0.0):

        return float(self.get(name, default))

    def getlist(self, name, default=None):

        value = self.get(name, default or [])
        if isinstance(value, str):
            value = value.split(',')
        return list(value)

    def getdict(self, name, default=None):

        value = self.get(name, default or {})
        if isinstance(value, str):
            value = json.loads(value)
        return dict(value)

    def getwithbase(self, name):

        compbs = BaseSettings()
        compbs.update(self[name + '_BASE'])
        compbs.update(self[name])
        return compbs

    def getpriority(self, name):

        if name not in self:
            return None
        return self.attributes[name].priority

    def maxpriority(self):

        if len(self) > 0:
            return max(self.getpriority(name) for name in self)
        else:
            return get_settings_priority('default')

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, priority='project'):
        # 设置存储键值对
        self._assert_mutability()
        # 把优先级名称转换为数值
        priority = get_settings_priority(priority)
        if name not in self:
            if isinstance(value, SettingsAttribute):
                self.attributes[name] = value
            else:
                # 用value实例化一个SettingsAttribute
                self.attributes[name] = SettingsAttribute(value, priority)
        else:
            self.attributes[name].set(value, priority)

    def setdict(self, values, priority='project'):
        self.update(values, priority)

    # 传入一个模块对象，或者模块的导入路径
    def setmodule(self, module, priority='project'):
        # 将一个模块里面的配置项存储起来
        self._assert_mutability()
        if isinstance(module, str):
            module = import_module(module)
        for key in dir(module):
            # 只有大写的配置项名称才能被存储起来
            if key.isupper():
                self.set(key, getattr(module, key), priority)

    def update(self, values, priority='project'):
        # 判断是否可修改
        self._assert_mutability()
        if isinstance(values, str):
            # 如果values是字符串类型，则序列化陈json对象
            values = json.loads(values)
        if values is not None:
            if isinstance(values, BaseSettings):
                # six.iteritems返回可迭代对象
                for name, value in values.items():
                    self.set(name, value, values.getpriority(name))
            else:
                for name, value in values.items():
                    self.set(name, value, priority)

    def delete(self, name, priority='project'):
        self._assert_mutability()
        priority = get_settings_priority(priority)
        if priority >= self.getpriority(name):
            del self.attributes[name]

    def __delitem__(self, name):
        self._assert_mutability()
        del self.attributes[name]

    def _assert_mutability(self):
        if self.frozen:
            raise TypeError("Trying to modify an immutable Settings object")

    def copy(self):
        return copy.deepcopy(self)

    def freeze(self):
        self.frozen = True

    def frozencopy(self):
        copy = self.copy()
        copy.freeze()
        return copy

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)

    def _to_dict(self):
        return {k: (v._to_dict() if isinstance(v, BaseSettings) else v)
                for k, v in self.items()}

    def copy_to_dict(self):

        settings = self.copy()
        return settings._to_dict()

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(repr(self))
        else:
            p.text(pformat(self.copy_to_dict()))


class _DictProxy(MutableMapping):

    def __init__(self, settings, priority):
        self.o = {}
        self.settings = settings
        self.priority = priority

    def __len__(self):
        return len(self.o)

    def __getitem__(self, k):
        return self.o[k]

    def __setitem__(self, k, v):
        self.settings.set(k, v, priority=self.priority)
        self.o[k] = v

    def __delitem__(self, k):
        del self.o[k]

    def __iter__(self, k, v):
        return iter(self.o)


class Settings(BaseSettings):

    def __init__(self, values=None, priority='project'):
        super(Settings, self).__init__()
        # 配置对象加载的时候，优先加载项目的默认配置
        self.setmodule(default_settings, 'default')
        for name, val in self.items():
            # 把配置项是字典的单独拿出来
            if isinstance(val, dict):
                self.set(name, BaseSettings(val, 'default'), 'default')
        self.update(values, priority)


def iter_default_settings():
    for name in dir(default_settings):
        if name.isupper():
            return name, getattr(default_settings, name)


def overridden_settings(settings):
    for name, defvalue in iter_default_settings():
        value = settings[name]
        if not isinstance(defvalue, dict) and value != defvalue:
            return name, value



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/16 19:42'


from collections import defaultdict
from typing import  Iterator, Dict


class Cookie:
    options = {
        'samesite': ('same_site', lambda x: True),
        'httponly': ('http_only', lambda x: True),
        'expires': ('expires_at', lambda x: x[1]),
        'domain': ('domain', lambda x: x[1]),
        'path': ('path', lambda x: x[1]),
        'max-age': ('max_age', lambda x: int(x[1])),
        'secure': ('secure', lambda x: True)
    }
    __slots__ = ('name', 'value', 'same_site', 'http_only', 'expires_at',
                 'path', 'domain', 'max_age', 'secure')

    def __init__(self, name: str, value: str = None, same_site=False,
                 http_only=False, expires_at=None, path: str = None,
                 domain: str = None, max_age: int = None, secure: bool = None):
        self.name = name
        self.value = value if value is not None else ''
        self.same_site = same_site
        self.http_only = http_only
        self.expires_at = expires_at
        self.path = path
        self.domain = domain
        self.max_age = max_age
        self.secure = secure

    @property
    def header(self):
        header = f'Set-Cookie: {self.name}={self.value}; '
        if self.same_site:
            header += 'SameSite; '
        if self.http_only:
            header += 'HttpOnly; '
        if self.expires_at:
            header += f'Expires={self.expires_at}; '
        if self.domain:
            header += f'Domain={self.domain}; '
        if self.path:
            header += f'Path={self.path}; '
        return header.encode()

    @classmethod
    def from_header(cls, header: str):
        instance = cls(name='')
        first = True
        for piece in header.split(';'):
            sub_pieces = piece.split('=')
            if sub_pieces[0] == '':
                continue
            if first:
                instance.name = sub_pieces[0]
                if len(sub_pieces) == 2:
                    instance.value = sub_pieces[1]
                first = False
            else:
                try:
                    attribute_name, value = cls.options[sub_pieces[0].lower().strip()]
                except KeyError as error:
                    pass
                setattr(instance, attribute_name, value(sub_pieces))
        return instance


class CookiesJar:
    __slots__ = ('cookies',)

    def __init__(self, seq:Dict=None):
        self.cookies = {}
        if seq is not None:
            self.cookies.update(seq)

    def add_cookie(self, cookie: Cookie):
        self.cookies[cookie.name] = cookie

    def get(self, name: str):
        return self.cookies.get(name)

    def update(self, seq: Dict):
        self.cookies.update(seq)

    def __getitem__(self, item):
        return self.cookies[item]

    def __setitem__(self, key, value):
        if not isinstance(value, Cookie):
            raise Exception('Value must be a Cookie instance.')
        self.cookies[key] = value

    def __iter__(self):
        return self.cookies.values().__iter__()

    def merge(self, newest_jar):
        for cookie_name, cookie in newest_jar.cookies.items():
            self.cookies[cookie_name] = cookie

    def __bool__(self):
        return bool(self.cookies)

    def __str__(self):
        return '<CookiesJar(' + str([x for x in self.cookies.items()]) + ')>'


class SessionCookiesJar:
    __slots__ = ('_domains',)

    def __init__(self):
        self._domains = defaultdict(CookiesJar)

    def get(self, domain: str, strict: bool = False) -> CookiesJar:
        if not strict:
            unique_jar = CookiesJar()
            for key, jar in self._domains.items():
                if key.endswith(domain):
                    unique_jar.merge(jar)
            return unique_jar
        return self._domains.get(domain)

    def add_cookie(self, cookie: Cookie):
        self._domains[cookie.domain].add_cookie(cookie)

    @property
    def cookies(self) -> Iterator:
        for jar in self._domains.values():
            for cookie in jar.cookies.values():
                yield cookie

    def merge(self, jar: CookiesJar, domain: str):
        self._domains[domain].merge(jar)



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-19 23:28'

# 项目名称
PROJECT_NAME = "spirit"

SERVER_IP = "127.0.0.1"

SERVER_PORT = 8088

# 同时并发的请求数量
WORKER_NUM = 3

# 记录日志的文件的path，如果为空则，不保存日志。直接输出
LOG_FILE_PATH = ""

# download
# 运行同时执行的下载任务的数量
MAX_DOWNLOAD_NUM = 3
# 请求的默认http头部
HEADERS = {
    'Accept': 'application/json, text/javascript, */*; q=0.01',
    'Accept-Encoding': 'gzip, deflate',
    'Accept-Language': 'zh-CN,zh;q=0.9',
    'Connection': 'keep-alive',
    'User-Agent': "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.86 Safari/537.36",
    'X-Requested-With': 'XMLHttpRequest',
}
# 是否运行跳转
FOLLOW_REDIRECTS = False
# 最大跳转次数
MAX_REDIRECTS = 3
# 是否是流式数据
STREAM = False

DECODE = True

SSL = None
# 是否长连接
KEEP_ALIVE = True
#
PREFIX = ""
# 超时时间
TIMEOUT = 5
# 每一个下载请求之间是的延迟时间
DELAY = 1
# 限流的策略
LIMITS = None

# schedule
# 布隆过滤器的容量
INITIAL_CAPACITY = 100000
# 布隆过滤器的错误率
ERROR_RATE = 0.001
# 队列中的请求是否长期保存
LONG_SAVE = False
# 队列中的请求若长期保存的话，需指定保存的目录
JOB_DIR = ""

# PyCoroutinePool
# 协程池中最大运行的协程数量
MAXCOROUTINEAMOUNT = 256 * 1024
# 协程的空闲时间
MAXCOROUTINEIDLETIME = 10
# 每一个协程运行完毕之后是否立即退出，
MOSTSTOP = True

ITEM_COUNTER_UNIT = 60

RESPONSE_COUNTER_UNIT = 60



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-24 12:50'


import time
from sprite import Downloader
from sprite import Request
from sprite import PyCoroutinePool


headers = {'Accept': 'text/html, application/xhtml+xml, image/jxr, */*',
           'Accept - Encoding': 'gzip, deflate',
           'Accept-Language': 'zh-Hans-CN, zh-Hans; q=0.5',
           'Connection': 'Keep-Alive',
           'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Safari/537.36 Edge/15.15063'}


def parse(response):
    pass


if __name__ == '__main__':
    url = "https://www.bilibili.com/index/recommend.json"
    request = Request(url=url, headers=headers, callback=parse,
                      meta={"proxy": "http://106.42.211.175:30007"}
                      )
    coroutine_pool = PyCoroutinePool()
    downloader = Downloader(
        coroutine_pool=coroutine_pool, loop=coroutine_pool.loop)
    coroutine_pool.start()
    start_time = time.time()
    coroutine_pool.go(downloader.request(request=request))
    print(f'等待获取下载好的response {time.time()-start_time}')
    response = downloader.getResponse()
    print(response.body)
    print(f'关闭下载器 {time.time()-start_time}')
    downloader.close()
    print(f'关闭协程池 {time.time() - start_time}')
    coroutine_pool.stop()



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019/8/27 19:55'

import asyncio
import socket
from contextlib import suppress


def tcp_nodelay(transport: asyncio.Transport, value: bool) -> None:
    sock = transport.get_extra_info('socket')

    if sock is None:
        return

    if sock.family not in (socket.AF_INET, socket.AF_INET6):
        return

    value = bool(value)

    # socket may be closed already, on windows OSError get raised
    with suppress(OSError):
        sock.setsockopt(
            socket.IPPROTO_TCP, socket.TCP_NODELAY, value)



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-17 22:47'

from sprite.utils.log import get_logger


class Spider:
    name = "sprite"
    start_requests: list = None
    metadata: dict = {}

    def __init__(self):
        self.logger = get_logger()

    async def start_request(self):
        pass

    async def parse(self, response: "Response"):
        pass



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-21 00:47'


import re
import time
import asyncio


class RequestRate:
    __slots__ = ('_count', '_period', '_actual_usage', '_next_flush', 'pattern', 'optimistic')

    def __init__(self, count:int, period:int, pattern=None, optimistic:bool=False):
        """

        :param count: How many requests per period are allowed.
                              If the request rate is 10 requests per minute than count=10 with period=60.
        :param period: How many seconds between each cycle.
                               If the request rate is 10 requests per minute than the period is 60.
        :param pattern: A pattern to match against URLs and skip the rate for those that don't match the pattern.
        :param optimistic: Request rate is complicated because you never know how __exactly__ the server is measuring
        the throughput... By default frame is optimistic and assumes that the server will honor the rate/period
        accurately.
        """
        if isinstance(pattern, str):
            # 如果pattern是正则字符串
            self.pattern =re.compile(pattern)
        else:
            self.pattern = pattern
        self.optimistic = optimistic
        self._count=count
        self._period = period
        self._actual_usage = 0
        self._next_flush = None

    async def notify(self):
        # 请求认证通过了，可以接着执行下一步的代码
        now = time.time()
        if self._next_flush is None:
            # 下一次刷新时间戳没有确定，则是开始计时状态
            self._next_flush = now+self._period
        elif now >self._next_flush:
            # 目前时间已经到达下一次时间戳
            self._next_flush = now +self._period
            # 晴空 已经通过的请求数量
            self._actual_usage = 0
        elif now<self._next_flush and self._actual_usage>=self._count:
            # 还未到下一次重置已经通过请求数量，但是已经通过的请求数量已经到达阈值
            if self.optimistic:
                wait_time = (self._next_flush-now) or 0
            else:
                wait_time = self._period
            # 目前协程进行休眠指定的时间，以防止请求速度放行过快
            await asyncio.sleep(wait_time)
            # 到达下一次循环时间戳，则重置通过的请求的数量统计
            self._actual_usage = 0
        # 在一个周期内，允许已经通过的请求数量+1
        self._actual_usage +=1




# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-21 17:38'


class RetryStrategy:
    __slots__ = ('network_failures', 'responses')

    def __init__(self, network_failures: dict = None, responses: dict = None):
        self.network_failures = network_failures or {'GET': 1}

        self.responses = responses or {}

    def clone(self) -> 'RetryStrategy':
        return RetryStrategy(network_failures=self.network_failures.copy(), responses=self.responses.copy())



import sys
from setuptools import setup, find_packages


PY_VER = sys.version_info

if PY_VER < (3, 6):
    raise RuntimeError("Sprite doesn't support Python version prior 3.6")


setup(
    name="spriterx",
    version="0.1.1",
    author="liyong",
    description="Python 3.6+ web scraping micro-framework based on asyncio coroutine pool.",
    long_description_content_type="text/markdown",
    url="https://github.com/ttloveyy/sprite",
    author_email="819078740@qq.com",
    python_requires=">=3.6",
    install_requires=["httptools", "bitarray", "w3lib", "typing"],
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: POSIX :: BSD",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Source": "https://github.com/ttloveyy/sprite",
    },
)



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-17 22:47'


import signal

from sprite.utils.log import get_logger, set_logger
from sprite.settings import Settings
from sprite.middlewaremanager import MiddlewareManager
from sprite.spider import Spider
from sprite.core.engine import Engine


logger = get_logger()


class Crawler:
    def __init__(self, spider: Spider, middlewareManager: MiddlewareManager=None, settings: Settings = None, server: str = None):
        self.spider = spider
        self._settings = settings or Settings()
        self._settings.freeze()
        self._engine = Engine.from_settings(
            settings=self._settings, spider=self.spider, middlewareManager=middlewareManager)
        self._server = server or f'{self._settings.get("SERVER_IP")}:{self._settings.getint("SERVER_PORT")}'

    def run(self):
        set_logger(self._settings)
        # self._register_crawl()
        logger.info(f'启动sprite')
        self._engine.start()
        signal.signal(signal.SIGTERM, self._close)  # SIGTERM 关闭程序信号
        signal.signal(signal.SIGINT, self._close)  # 接收ctrl+c 信号

    def _close(self, signal_num: int, frame):
        # 关闭引擎
        if self._engine.isClose():
            return
        logger.info(f'关闭sprite')
        self._engine.close()

    # def _register_crawl(self):
    #     client_call(self._server,"register_crawl", self)


# class CrawlerRunner:
#     def __init__(self):
#         self._crawlers = set
#         self._settings = None
#         self._rpc_server = SpritRPCServer()
#         self._running_event = threading.Event()
#         self._register_rpc()
#
#     def _register_rpc(self):
#         self._rpc_server.register_function(self.register_crawl, "register_crawl")
#
#
#     def register_crawl(self, crawler:Crawler):
#         if crawler.spider.name in self._crawlers:
#             return False
#         self._crawlers[crawler.spider.name] =crawler
#         return True
#
#     # 重新加载settings
#     def load_settings(self):
#         new_settings = import_settings()
#         self.settings = new_settings
#
#     def start(self):
#         self._running_event.wait()  # 主线程阻塞等待
#         self._rpc_server.shutdown()  # 关闭rpc服务
#
#
#     def close_crawl(self, crawl_name):
#         if crawl_name in self._crawlers:
#             pass
#
#
#
#
#
# # 导入当前目录下的setting模块！！！！，找不到会直接抛出异常！！
# def import_settings(priority: str = 'project'):
#     settings_ = Settings()
#     try:
#         import settings
#         for key in dir(settings):
#             # 只有大写的配置项名称才能被存储起来
#             if key.isupper():
#                 settings_.set(key, getattr(settings, key), priority)
#         return settings_
#     except ImportError as e:
#         logger.info(f'import settings failure, not start crawl')
#         raise e



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-08-17 12:23'


import asyncio
from sprite.core.download.session import Session


headers = {'Accept': 'text/html, application/xhtml+xml, image/jxr, */*',
               'Accept - Encoding': 'gzip, deflate',
               'Accept-Language': 'zh-Hans-CN, zh-Hans; q=0.5',
               'Connection': 'Keep-Alive',
               'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Safari/537.36 Edge/15.15063'}



async def get_http_data(url:str):
    s = Session()
    resp = await s.get(url, headers=headers)
    print(resp.content.decode("utf-8")[:10])
    # print(resp.json())

def parse(response):
    pass



if __name__ == '__main__':
    url = "https://www.bilibili.com/index/recommend.json"
    # request = Request(url=url, headers=headers, callback=parse)
    # coroutine_pool = PyCoroutinePool()
    # downloader = Downloader(coroutine_pool=coroutine_pool,loop=coroutine_pool.loop)
    # coroutine_pool.stop()
    # coroutine_pool.go(downloader.request(request=request))
    # # resp = requests.get(url, headers=headers)
    # # print(resp.text)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(get_http_data(url))



# -*- coding:utf-8 -*-
__author__ = 'liyong'
__date__ = '2019-07-20 23:39'


import ssl
from asyncio import StreamReader,StreamWriter,AbstractEventLoop, wait_for, TimeoutError
from typing import Coroutine

SECURE_CONTEXT = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
SECURE_CONTEXT.check_hostname = True

INSECURE_CONTEXT = ssl.SSLContext()
INSECURE_CONTEXT.check_hostname = False


class Connection:
    __slots__ = ('loop','reader','writer', 'pool')
    def __init__(self, loop:AbstractEventLoop, reader:StreamReader,writer:StreamWriter,pool):
        self.loop = loop
        self.reader = reader
        self.writer = writer
        self.pool = pool # 一个域名的连接池

    # 发送数据，生成一个
    def sendall(self, data:bytes)->Coroutine:
        # print(f'send data : {data.decode("utf-8")}')
        self.writer.write(data)
        # 返回一个写数据协程，等待知道写到缓冲区中
        return self.writer.drain()

    # 读取固定长度的数据，从可读对象中
    def read_exactly(self, length:int)->Coroutine:
        return self.reader.readexactly(length)

    # 从可读对象中读取到指定的字节为止
    def read_until(self, delimiter:bytes)->Coroutine:
        return self.reader.readuntil(delimiter)

    def close(self):
        self.writer.close()

    async def is_dropped(self):
        try:
            # 验证连接是否空闲
            await wait_for(self.reader.readexactly(0), 0.001)
            return True
        except TimeoutError:
            # 捕获到超时异常
            return False

    def release(self, keep_alive:bool=False):
        # 把连接实例传入进去，关闭连接
        self.pool.release_connection(self, keep_alive)
